r'''
The 1D advection-diffusion problem
----------------------------------

The equation to be solved is

.. math::

    \frac{\partial}{\partial t} \psi(x,t) &= -\frac{1}{w(x)} \frac{\partial}{\partial x} \left[ w(x) ~ \mathcal{F}(x,t) \right] + \dot{\psi}\\
    \mathcal{F} &= U(x) \psi(x) -K(x) ~ \frac{\partial \psi}{\partial x} + F(x)

for the following quantities:

- state variable :math:`\psi(x,t)`
- diffusivity :math:`K(x)` in units of :math:`x^2 ~ t^{-1}`
- advecting velocity :math:`U(x)` in units of :math:`x ~ t^{-1}`
- a prescribed flux :math:`F(x)` (including boundary conditions) in units of :math:`\psi ~ x ~ t^{-1}`
- a scalar source/sink :math:`\dot{\psi}(x)` in units of :math:`\psi ~ t^{-1}`
- weighting function :math:`w(x)` for the divergence operator on curvilinear grids.

The boundary condition is a flux condition at the end points:

.. math::
    \begin{align}  \label{eq:fluxcondition}
    \mathcal{F}(x_0) &= F(x_0)   &    \mathcal{F}(x_J) &= F(x_J)
    \end{align}

which requires that the advecting velocity :math:`u(x) = 0` at the end points :math:`x_0, x_J`

The solver is implemented on a 1D staggered grid, with J+1 flux points
and J scalar points located somewhere between the flux points.

The solver does **not** assume the gridpoints are evenly spaced in :math:`x`.

Routines are provided to compute the following:

- Advective, diffusive, and total fluxes (the terms of :math:`\mathcal{F}`)
- Tridiagonal matrix operator for the flux convergence
- The actual flux convergence, or instantaneous scalar tendency given a current value of :math:`\psi(x)`
- Future value of :math:`\psi(x)` for an implicit timestep

Some details of the solver formulas are laid out below for reference.

Spatial discretization
----------------------

We use a non-uniform staggered spatial grid with scalar :math:`\psi` evaluated at :math:`J` points,
and flux :math:`\mathcal{F}` evaluated at :math:`J+1` flux points.
The indexing will run from :math:`j=0` to :math:`j=J` for the flux points,
and :math:`i=0` to :math:`i=J-1` for the scalar points.
This notation is consistent with zero-indexed Python arrays.

We define the following arrays:

- :math:`\mathcal{X}_b[j]` is a length J+1 array defining the location of the flux points.
- :math:`\mathcal{X}[i]` is a length J array defining the location of the scalar points, where point :math:`\mathcal{X}[j]` is somewhere between :math:`\mathcal{X}_b[j]` and :math:`\mathcal{X}_b[j+1]` for all :math:`j<J`.
- :math:`\psi[i], \dot{\psi}[i]` are length J arrays defined on :math:`\mathcal{X}`.
- :math:`U[j], K[j], F[j]` are all arrays of length J+1 defined on :math:`\mathcal{X}_b`.
- The grid weights are similarly in arrays :math:`W_b[j], W[i]` respectively on :math:`\mathcal{X}_b`` and :math:`\mathcal{X}`.

Centered difference formulas for the flux
-----------------------------------------

We use centered differences in :math:`x` to discretize the spatial derivatives.
The diffusive component of the flux is thus

.. math::

    \begin{align*}
    \mathcal{F}_{diff}[j] &= - K[j] \frac{ \left( \psi[i] - \psi[i-1] \right) }{\left( \mathcal{X}[i] - \mathcal{X}[i-1] \right)}  &  j&=i=1,2,...,J-1
    \end{align*}

The diffusive flux is assumed to be zero at the boundaries.

The advective term requires an additional approximation since the scalar :math:`\psi` is not defined at the flux points.
We use a linear interpolation to the flux points:

.. math::

    \begin{align*}
    \psi_b[j] &\equiv \psi[i-1]  \left( \frac{\mathcal{X}[i] - \mathcal{X}_b[j]}{\mathcal{X}[i] - \mathcal{X}[i-1]} \right)  +   \psi[i]  \left(  \frac{ \mathcal{X}_b[j] - \mathcal{X}[i-1] }{\mathcal{X}[i] - \mathcal{X}[i-1]} \right)   &   j&=i=1,2,...,J-1
    \end{align*}

Note that for an evenly spaced grid, this reduces to the simple average :math:`\frac{1}{2} \left(  \psi[i-1] +  \psi[i] \right)`.

With this interpolation, the advective flux is approximated by

.. math::

    \begin{align*}
    \mathcal{F}_{adv}[j] &= \frac{U[j] }{\mathcal{X}[i] - \mathcal{X}[i-1]} \left(  \psi[i-1] (\mathcal{X}[i] - \mathcal{X}_b[j])   +   \psi[i] (\mathcal{X}_b[j] - \mathcal{X}[i-1])   \right) &   j&=i=1,2,...,J-1
    \end{align*}

The total flux away from the boundaries (after some recombining terms) is thus:

.. math::

    \mathcal{F}[j] = F[j] + \psi[i-1] \left( \frac{K[j] + U[j] (\mathcal{X}[i] - \mathcal{X}_b[j]) }{ \mathcal{X}[i] - \mathcal{X}[i-1] }   \right)  - \psi[i]  \left(  \frac{K[j] - U[j] (\mathcal{X}_b[j] - \mathcal{X}[i-1]) }{\mathcal{X}[i] - \mathcal{X}[i-1] } \right)

which is valid for j=i=1,2,...,J-1.

Centered difference formulas for the flux convergence
-----------------------------------------------------

Centered difference approximation of the flux convergence gives

.. math::

    \begin{align*}
    \frac{\partial }{\partial t} \psi[i] &= -\frac{ W_b[j+1] \mathcal{F}[j+1] - W_b[j] \mathcal{F}[j] }{W[i] ( \mathcal{X}_b[j+1] - \mathcal{X}_b[j] )}  + \dot{\psi}[i] & i&=j=0,1,...,J-1
    \end{align*}

The flux convergences are best expressed together in matrix form:

.. math::

    \begin{equation}
    \frac{\partial \boldsymbol{\psi}}{\partial t} = \boldsymbol{T} ~ \boldsymbol{\psi} + \boldsymbol{S}
    \end{equation}

where :math:`\boldsymbol{\psi}` is the :math:`J\times1` column vector,
:math:`\boldsymbol{S}` is a :math:`J\times1` column vector
representing the prescribed flux convergence and source terms, whose elements are

.. math::

    \begin{align}
    S[i] &= \frac{-W_b[j+1] F[j+1] + W_b[j] F[j]}{W[i] ( \mathcal{X}_b[j+1] - \mathcal{X}_b[j] )}  + \dot{\psi}[i]  &  i&=j=0,1,...,J-1
    \end{align}

and :math:`\boldsymbol{T}` is a :math:`J\times J` tridiagonal matrix:

.. math::

    \begin{equation}
    \boldsymbol{T} ~ \boldsymbol{\psi} = \left[\begin{array}{ccccccc} T_{m0} & T_{u1} & 0 & ... & 0 & 0 & 0 \\T_{l0} & T_{m1} & T_{u2} & ... & 0 & 0 & 0 \\ 0 & T_{l1} & T_{m2} & ... & 0 & 0 & 0 \\... & ... & ... & ... & ... & ... & ... \\0 & 0 & 0 & ... & T_{m(J-3)} & T_{u(J-2)} & 0 \\0 & 0 & 0 & ... & T_{l(J-3)} & T_{m(J-2)} & T_{u(J-1)} \\0 & 0 & 0 & ... & 0 & T_{l(J-2)} & T_{m(J-1)}\end{array}\right] \left[\begin{array}{c} \psi_0 \\ \psi_1 \\ \psi_2 \\... \\ \psi_{J-3} \\ \psi_{J-2} \\ \psi_{J-1} \end{array}\right]
    \end{equation}

with vectors :math:`T_l, T_m, T_u` representing respectively the lower, main, and upper diagonals of :math:`\boldsymbol{T}`.
We will treat all three vectors as length J;
the 0th element of :math:`T_u` is ignored while the (J-1)th element of :math:`T_l` is ignored
(this is consistent with the expected inputs for the Python module scipy.linalg.solve_banded).

The instantanous tendency is then easily computed by matrix multiplication.

The elements of the main diagonal of :math:`\boldsymbol{\psi}` can be computed from

.. math::

    \begin{align}  \label{eq:maindiag}
    \begin{split}
    T_m[i] &= -\left( \frac{ W_b[j+1]  \big( K[j+1] + U[j+1] (\mathcal{X}[i+1] - \mathcal{X}_b[j+1]) \big) }{ W[i] ( \mathcal{X}_b[j+1] - \mathcal{X}_b[j] )(\mathcal{X}[i+1] - \mathcal{X}[i]) }   \right)  \\
    & \qquad - \left(  \frac{W_b[j] \big( K[j] - U[j] (\mathcal{X}_b[j] - \mathcal{X}[i-1]) \big) }{W[i] ( \mathcal{X}_b[j+1] - \mathcal{X}_b[j] )(\mathcal{X}[i] - \mathcal{X}[i-1]) } \right)  \\
    i &=j=0,2,...,J-1
    \end{split}
    \end{align}

which is valid at the boundaries so long as we set :math:`W_b[0] = W_b[J] = 0`.

The lower diagonal (including the right boundary condition) is computed from

.. math::

    \begin{align}  \label{eq:lowerdiag}
    \begin{split}
    T_l[i-1] &= \left( \frac{W_b[j]}{W[i] } \right) \left( \frac{ K[j] + U[j] (\mathcal{X}[i] - \mathcal{X}_b[j]) }{( \mathcal{X}_b[j+1] - \mathcal{X}_b[j] ) (\mathcal{X}[i] - \mathcal{X}[i-1] )}   \right) \\
     i &=j =1,2,...,J-2, J-1
     \end{split}
    \end{align}

Finally the upper diagonal (including the left boundary condition) is computed from

.. math::

    \begin{align}  \label{eq:upperdiag}
    \begin{split}
    T_u[i+1] &= \left( \frac{W_b[j+1]}{W[i]} \right) \left( \frac{K[j+1] - U[j+1] (\mathcal{X}_b[j+1] - \mathcal{X}[i]) }{( \mathcal{X}_b[j+1] - \mathcal{X}_b[j] )(\mathcal{X}[i+1] - \mathcal{X}[i] ) }  \right)  \\
    i &= j=0,...,J-2
    \end{split}
    \end{align}

Implicit time discretization
----------------------------

The forward-time finite difference approximation to LHS of the flux-convergence equation is simply

.. math::
    \begin{equation}
    \frac{\partial \psi[i]}{\partial t} \approx \frac{\psi^{n+1}[i]- \psi^{n}[i]}{\Delta t}
    \end{equation}

where the superscript :math:`n` indicates the time index.

We use the implicit-time method, in which the RHS is evaluated at the future time :math:`n+1`.
Applying this to the matrix equation above
and moving all the terms at time :math:`n+1` over to the LHS yields

.. math::

    \begin{equation}  \label{eq:implicit_tridiagonal}
    \left( \boldsymbol{I} - \boldsymbol{T} \Delta t \right) \boldsymbol{\psi}^{n+1} = \boldsymbol{\psi}^{n} + \boldsymbol{S} \Delta t
    \end{equation}

where :math:`\boldsymbol{I}` is the :math:`J\times J` identity matrix.

Solving for the future value :math:`\boldsymbol{\psi}^{n+1}` is then accomplished
by solving the :math:`J \times J` tridiagonal linear system using standard routines.

Analytical benchmark
--------------------

Here is an analytical case to be used for testing purposes to validate the numerical code.
This is implemented in the CLIMLAB test suite.

- :math:`K=K_0` is constant
- :math:`w(x) = 1` everywhere (Cartesian coordinates)
- :math:`F = 0` everywhere
- :math:`\psi(x,0) = \psi_0 \sin^2\left(\frac{\pi x}{L}\right)`
- :math:`u(x) = U_0 \sin\left(\frac{\pi x}{L}\right)`
for a domain with endpoints at :math:`x=0` and :math:`x=L`.

The analytical solution is

.. math::

    \begin{align}
    \mathcal{F} &= \psi_0 \sin\left(\frac{\pi x}{L}\right) \left[U_0 \sin^2\left(\frac{\pi x}{L}\right) - 2K \frac{\pi}{L} \cos\left(\frac{\pi x}{L}\right) \right]  \\
    \frac{\partial \psi}{\partial t} &= -\psi_0 \frac{\pi}{L} \left\{ 3 U_0 \sin^2\left(\frac{\pi x}{L}\right) \cos\left(\frac{\pi x}{L}\right) -2K\frac{\pi}{L} \left[\cos^2\left(\frac{\pi x}{L}\right) -\sin^2\left(\frac{\pi x}{L}\right) \right] \right\}
    \end{align}

which satisfies the boundary condition :math:`\mathcal{F} = 0` at :math:`x=0` and :math:`x=L`.


Module function reference
-------------------------

All the functions in ``climlab.dynamics.adv_diff_numerics`` are vectorized
to handle multidimensional input. The key assumption is that
**advection-diffusion operates along the final dimension**.

Inputs should be reshaped appropriately (e.g. with ``numpy.moveaxis()``)
before calling these functions.
'''
from __future__ import division
from numpy import zeros, ones, zeros_like, ones_like, matmul, diag, diag_indices, diff, newaxis
from numpy.linalg import solve
from scipy.linalg import solve_banded

def diffusive_flux(X, Xb, K, field):
    '''Return the diffusive flux on cell boundaries (length J+1)'''
    flux = zeros_like(K)
    flux[...,1:-1] += field[...,:-1]*K[...,1:-1]/diff(X,axis=-1)
    flux[...,1:-1] -= field[...,1:]*K[...,1:-1]/diff(X,axis=-1)
    return flux

def advective_flux(X, Xb, U, field):
    '''Return the advective flux on cell boundaries (length J+1)'''
    flux = zeros_like(U)
    flux[...,1:-1] += field[...,:-1]*(U[...,1:-1]*(X[...,1:]-Xb[...,1:-1]))/diff(X,axis=-1)
    flux[...,1:-1] -= field[...,1:]*(-U[...,1:-1]*(Xb[...,1:-1]-X[...,:-1]))/diff(X,axis=-1)
    return flux

def total_flux(X, Xb, K, U, field, prescribed_flux=None):
    '''Return the total (advective + diffusive + prescribed) flux
       on cell boundaries (length J+1)'''
    if prescribed_flux is None:
        prescribed_flux = zeros_like(U)
    return advective_flux(X, Xb, U, field) + diffusive_flux(X, Xb, K, field) + prescribed_flux

def advdiff_tridiag(X, Xb, K, U, W=None, Wb=None, use_banded_solver=False):
    r'''Compute the tridiagonal matrix operator for the advective-diffusive
       flux convergence.

       Input arrays of length J+1:
         Xb, Wb, K, U
       Input arrays of length J:
         X, W

       The 0th and Jth (i.e. first and last) elements of Wb are ignored;
       assuming boundary condition is a prescribed flux.

       The return value depends on input flag ``use_banded_solver``

       If ``use_banded_solver==True``, return a 3xJ array containing the elements of the tridiagonal.
       This version is restricted to 1D input arrays,
       but is suitable for use with the efficient banded solver.

       If ``use_banded_solver=False`` (which it must be for multidimensional input),
       return an array (...,J,J) with the full tridiagonal matrix.
    '''
    J = X.shape[-1]
    if (W is None):
        W = ones_like(X)
    if (Wb is None):
        Wb = ones_like(Xb)
    #  These are all length (J-1) in the last axis
    lower_diagonal = (Wb[...,1:-1]/W[...,1:] *
        (K[...,1:-1]+U[...,1:-1]*(X[...,1:]-Xb[...,1:-1])) /
        ((Xb[...,2:]-Xb[...,1:-1])*(X[...,1:]-X[...,:-1])))
    upper_diagonal = (Wb[...,1:-1]/W[...,:-1] *
        (K[...,1:-1]-U[...,1:-1]*(Xb[...,1:-1]-X[...,:-1])) /
        ((Xb[...,1:-1]-Xb[...,:-2])*(X[...,1:]-X[...,:-1])))
    main_diagonal_term1 = (-Wb[...,1:-1]/W[...,:-1] *
        (K[...,1:-1]+U[...,1:-1]*(X[...,1:]-Xb[...,1:-1])) /
        ((Xb[...,1:-1]-Xb[...,:-2])*(X[...,1:]-X[...,:-1])))
    main_diagonal_term2 = (-Wb[...,1:-1]/W[...,1:] *
        (K[...,1:-1]-U[...,1:-1]*(Xb[...,1:-1]-X[...,:-1])) /
        ((Xb[...,2:]-Xb[...,1:-1])*(X[...,1:]-X[...,:-1])))
    if use_banded_solver:
        # Pack the diagonals into a 3xJ array
        tridiag_banded = zeros((3,J))
        # Lower diagonal (last element ignored)
        tridiag_banded[2,:-1] = lower_diagonal
        # Upper diagonal (first element ignored)
        tridiag_banded[0,1:] = upper_diagonal
        # Main diagonal, term 1, length J-1
        tridiag_banded[1,:-1] += main_diagonal_term1
        # Main diagonal, term 2, length J-1
        tridiag_banded[1, 1:] += main_diagonal_term2
        return tridiag_banded
    else:
        #  If X.size is (...,J), then the tridiagonal operator is (...,J,J)
        sizeJJ = tuple([n for n in X.shape[:-1]] + [J,J])
        tridiag = zeros(sizeJJ)
        #  indices for main, upper, and lower diagonals of a JxJ matrix
        inds_main = diag_indices(J)
        inds_upper = (inds_main[0][:-1], inds_main[1][1:])
        inds_lower = (inds_main[0][1:], inds_main[1][:-1])
        # Lower diagonal (length J-1)
        tridiag[...,inds_lower[0],inds_lower[1]] = lower_diagonal
        # Upper diagonal (length J-1)
        tridiag[...,inds_upper[0],inds_upper[1]] = upper_diagonal
        # Main diagonal, term 1, length J-1
        tridiag[...,inds_main[0][:-1],inds_main[1][:-1]] += main_diagonal_term1
        # Main diagonal, term 2, length J-1
        tridiag[...,inds_main[0][1:],inds_main[1][1:]] += main_diagonal_term2
        return tridiag

def make_the_actual_tridiagonal_matrix(tridiag_banded):
    '''Convert a (3xJ) banded array into full (JxJ) tridiagonal matrix form.'''
    return (diag(tridiag_banded[1,:], k=0) +
            diag(tridiag_banded[0,1:], k=1) +
            diag(tridiag_banded[2,:-1], k=-1))

def compute_source(X, Xb, prescribed_flux=None, prescribed_source=None,
                   W=None, Wb=None):
    '''Return the source array S consisting of the convergence of the prescribed flux
    plus the prescribed scalar source.'''
    if (W is None):
        W = ones_like(X)
    if (Wb is None):
        Wb = ones_like(Xb)
    if prescribed_flux is None:
        prescribed_flux = zeros_like(Xb)
    if prescribed_source is None:
        prescribed_source = zeros_like(X)
    F = prescribed_flux
    return ((-Wb[...,1:]*F[...,1:]+Wb[...,:-1]*F[...,:-1]) /
            (W*(Xb[...,1:]-Xb[...,:-1])) + prescribed_source)

def compute_tendency(field, tridiag, source, use_banded_solver=False):
    r'''Return the instantaneous scalar tendency.

    This is the sum of the convergence of advective+diffusive flux plus any
    prescribed convergence or scalar sources.

    The convergence is computed by matrix multiplication:

    .. math::

        \frac{\partial \psi}{\partial t} = T \times \psi + S

    where :math:`T` is the tridiagonal flux convergence matrix.
    '''
    if use_banded_solver:
        tridiag = make_the_actual_tridiagonal_matrix(tridiag)
    #  np.matmul expects the final 2 dims of each array to be matrices
    #  add a singleton dimension to field so we get (J,J)x(J,1)->(J,1)
    result = matmul(tridiag, field[...,newaxis]) + source[...,newaxis]
    #  Now strip the extra dim
    return result[...,0]

def implicit_step_forward(initial_field, tridiag, source, timestep,
                          use_banded_solver=False):
    r'''Return the field at future time using an implicit timestep.

    The matrix problem is

    .. math::

        (I - T \Delta t) \psi^{n+1} = \psi^n + S \Delta t

    where :math:`T` is the tridiagonal matrix for the flux convergence, :math:`psi` is the
    state variable, the superscript :math:`n` refers to the time index, and :math:`S \Delta t`
    is the accumulated source over the timestep :math:`\Delta t`.

    Input arguments:

    - ``initial_field``: the current state variable :math:`\psi^n`, dimensions (...,J)
    - ``tridiag``: the tridiagonal matrix :math:`T`, dimensions (...,J,J) or (...,3,J) depending on the value of ``use_banded_solver``
    - ``source``: prescribed sources/sinks of :math:`\psi`, dimensions (...,J)
    - ``timestep``: the discrete timestep in time units
    - ``use_banded_solver``: switch to use the optional efficient banded solver (see below)

    Returns the updated value of the state variable :math:`\psi^{n+1}`, dimensions (...,J)

    The expected shape of ``tridiag`` depends on the switch ``use_banded_solver``,
    which should be consistent with that used in the call to ``advdiff_tridiag()``.
    If ``True``, we use the efficient banded matrix solver
    ``scipy.linalg.solve_banded()``.
    However this will probably only work for a 1D state variable.

    The default is to use the general linear system solver ``numpy.linalg.solve()``.
    '''
    RHS = initial_field + source*timestep
    I = 0.*tridiag
    J = initial_field.shape[-1]
    if use_banded_solver:
        I[1,:] = 1.  # identity matrix in banded form
        IminusTdt = I-tridiag*timestep
        return solve_banded((1, 1), IminusTdt, RHS)
    else:
        #  indices for main, upper, and lower diagonals of a JxJ matrix
        inds_main = diag_indices(J)
        I = 0.*tridiag
        I[...,inds_main[0],inds_main[1]] = 1.  # stacked identity matrix
        IminusTdt = I-tridiag*timestep
        return solve(IminusTdt, RHS)



'''
Modules for simple dynamics, mostly for use in Energy Balance Models.

:class:`~climlab.dynamics.BudykoTransport` is a relaxation to global mean.

Other modules are 1D advection-diffusion solvers (implemented using implicit timestepping).

:class:`~climlab.dynamics.AdvectionDiffusion` is a general-purpose 1D
advection-diffusion process. It can be used out-of-the-box for models with
Cartesian grid geometry, but also accepts weighting functions for the
divergence operator on curvilinear grids.

:class:`~climlab.dynamics.MeridionalAdvectionDiffusion` implements the
1D advection-diffusion process on the sphere (flux in the north-south direction).

Subclass :class:`~climlab.dynamics.MeridionalHeatDiffusion` is the appropriate class
for the traditional diffusive EBM, in which transport is parameterized as a
meridional diffusion process down the zonal-mean surface temperature gradient.

:class:`~climlab.dynamics.MeridionalMoistDiffusion` implements the moist EBM,
with transport down an approximate gradient in near-surface moist static energy.
'''

from __future__ import absolute_import
from .budyko_transport import BudykoTransport
from .advection_diffusion import AdvectionDiffusion, Diffusion
from .meridional_advection_diffusion import MeridionalAdvectionDiffusion, MeridionalDiffusion
from .meridional_heat_diffusion import MeridionalHeatDiffusion
from .meridional_moist_diffusion import MeridionalMoistDiffusion



'''
Modules for self-describing gridded fields in climlab.
'''
__all__ = ['axis', 'domain', 'field', 'initial', 'xarray']

from climlab.domain.domain import single_column, zonal_mean_surface, surface_2D, zonal_mean_column, box_model_domain
from climlab.domain.initial import column_state, surface_state
from climlab.domain.field import Field, global_mean
from climlab.domain.axis import Axis



''' climlab wrapper for RRTMG radiation schemes.

This is implemented with classes :class:`~climlab.radiation.rrtm.RRTMG_LW`, and
:class:`~climlab.radiation.rrtm.RRTMG_SW`,
as well as a container class :class:`~climlab.radiation.rrtm.RRTMG`
that has LW and SW radiation as subprocesses.

Input arguments and diagnostics follow specifications in
:class:`~climlab.radiation._Radiation`

See <http://rtweb.aer.com/rrtm_frame.html> for more information about the RRTMG code.

    :Example:

        Here is a quick example of setting up a single-column
        Radiative-Convective model with fixed relative humidity::

            import climlab
            alb = 0.25
            #  State variables (Air and surface temperature)
            state = climlab.column_state(num_lev=30)
            #  Fixed relative humidity
            h2o = climlab.radiation.ManabeWaterVapor(name='WaterVapor', state=state)
            #  Couple water vapor to radiation
            rad = climlab.radiation.RRTMG(name='Radiation', state=state, specific_humidity=h2o.q, albedo=alb)
            #  Convective adjustment
            conv = climlab.convection.ConvectiveAdjustment(name='Convection', state=state, adj_lapse_rate=6.5)
            #  Couple everything together
            rcm = climlab.couple([rad,h2o,conv], name='Radiative-Convective Model')
            #  Run the model
            rcm.integrate_years(1)
            #  Check for energy balance
            print(rcm.ASR - rcm.OLR)

'''
from __future__ import division, absolute_import
from .rrtmg import RRTMG
from .rrtmg_lw import RRTMG_LW
from .rrtmg_sw import RRTMG_SW
from .utils import _climlab_to_rrtm, _rrtm_to_climlab



from __future__ import division
import climlab


def to_xarray(model):
    model.to_xarray()
    model.to_xarray(diagnostics=True)
    if hasattr(model, 'timeave'):
        climlab.to_xarray(model.timeave)
    # We should be able to render all tendencies as xarray
    for name, proc, top_proc in climlab.utils.walk.walk_processes(model, topdown=False):
        climlab.to_xarray(proc.tendencies)



import climlab
from climlab.utils import constants as const
import numpy as np
import matplotlib.pyplot as plt

# creating model
model = climlab.EBM_seasonal()
model.step_forward()

solar = model.subprocess['insolation'].insolation

# plot
fig = plt.figure( figsize=(6,4))
ax = fig.add_subplot(111)

season_days = const.days_per_year/4

for season in ['winter','spring','summer','autumn']:
    ax.plot(model.lat, solar, label=season)
    model.integrate_days(season_days)

ax.set_title('seasonal solar distribution')
ax.set_xlabel('latitude')
ax.set_xticks([-90,-60,-30,0,30,60,90])
ax.set_ylabel('solar insolation (W/m$^2$)')
ax.legend(loc='best')
plt.show()



from __future__ import division
from builtins import range
import numpy as np
from climlab.radiation.greygas import GreyGas
from climlab import constants as const
from climlab.domain import domain, axis, field
from copy import copy


class NbandRadiation(GreyGas):
    '''Process for radiative transfer.
    Solves the discretized Schwarschild two-stream equations
    with the spectrum divided into N spectral bands.

    Every NbandRadiation object has an attribute
    ``self.band_fraction``
    with sum(self.band_fraction) == 1
    that gives the fraction of the total beam in each band

    Also a dictionary
    ``self.absorber_vmr``
    that gives the volumetric mixing ratio of every absorbing gas
    on the same grid as temperature

    and a dictionary
    ``self.absorption_cross_section``
    that gives the absorption cross-section per unit mass for each gas
    in every spectral band
    '''
    def __init__(self, absorber_vmr=None, **kwargs):
        super(NbandRadiation, self).__init__(**kwargs)
        newinput = ['band_fraction',
                    'absorber_vmr',
                    'absorption_cross_section',
                    'cosZen',]
        self.declare_input(newinput)
        # this should be overridden by daughter classes
        self.band_fraction = np.array(1.)
        ##  a dictionary of absorbing gases, in volumetric mixing ratios
        #  each item should have dimensions of self.Tatm
        #  Can be passed as input argument
        if absorber_vmr is None:
            absorber_vmr = {}
        self.absorber_vmr = absorber_vmr
        # a dictionary of absorption cross-sections in m**2 / kg
        # each item should have dimension...  (num_channels, 1)
        self.absorption_cross_section = {}
        self.cosZen = 1.  # cosine of the average zenith angle
        dp = self.Tatm.domain.lev.delta
        self.mass_per_layer = dp * const.mb_to_Pa / const.g
        self.albedo_sfc = np.ones_like(self.band_fraction) * self.albedo_sfc

    @property
    def band_fraction(self):
        return self._band_fraction
    @band_fraction.setter
    def band_fraction(self, value):
        self.num_channels = value.size
        # abstract axis for channels
        ax = axis.Axis(num_points=self.num_channels)
        self.channel_ax = {'channel': ax}
        dom = domain._Domain(axes=self.channel_ax)
        #   fraction of the total solar flux in each band:
        self._band_fraction = field.Field(value, domain=dom)

    def _compute_optical_path(self):
        # this will cause a problem for a model without CO2
        tau = np.zeros_like(self.absorber_vmr['CO2']*
                            self.absorption_cross_section['CO2'])
        for gas, vmr in self.absorber_vmr.items():
            # convert to mass of absorber per unit total mass
            if gas == 'H2O':  # H2O is stored as specific humidity, not VMR
                q = vmr
            else:
                q = vmr / (1.+vmr)
            try:
                # if this gas isn't present in absorption dictionary
                # the assumption is that there is no absorption!
                kappa = self.absorption_cross_section[gas]
                tau += q * kappa
            except: pass
        tau *= self.mass_per_layer / self.cosZen
        return tau

    def _compute_absorptivity(self):
        #  assume that the water vapor etc is current
        optical_path = self._compute_optical_path()
        #  account for finite layer depth
        absorptivity = 1. - np.exp(-optical_path)
        axes = copy(self.Tatm.domain.axes)
        # add these to the dictionary of axes
        axes.update(self.channel_ax)
        dom = domain.Atmosphere(axes=axes)
        self.absorptivity = field.Field(absorptivity, domain=dom)

    def _compute_emission_sfc(self):
        #  need to split the total emission across the bands
        total_emission = super(NbandRadiation, self)._compute_emission_sfc()
        return self._split_channels(total_emission)

    def _compute_emission(self):
        #  need to split the total emission across the bands
        total_emission = super(NbandRadiation, self)._compute_emission()
        band_fraction = self.band_fraction
        for n in range(self.Tatm.domain.numdims):
            band_fraction = band_fraction[:, np.newaxis]
        return total_emission * band_fraction

    def _compute_radiative_heating(self):
        #  need to recompute transmissivities each time because
        # water vapor is changing
        self._compute_absorptivity()
        super(NbandRadiation, self)._compute_radiative_heating()

    def _split_channels(self, flux):
        split = np.outer(self.band_fraction, flux)
        # make sure there's a singleton dimension at the last axis (level)
        if np.size(split, axis=-1) != 1:
            split = split[..., np.newaxis]
        return split

    def _join_channels(self, flux):
        return np.sum(flux, axis=0)


class ThreeBandSW(NbandRadiation):
    def __init__(self, emissivity_sfc=0., **kwargs):
        '''A three-band mdoel for shortwave radiation.

        The spectral decomposition used here is largely based on the
        "Moist Radiative-Convective Model" by Aarnout van Delden, Utrecht University
        a.j.vandelden@uu.nl
        http://www.staff.science.uu.nl/~delde102/RCM.htm

        Three SW channels:
            channel 0 is Hartley and Huggins band (UV, 1%, 200 - 340 nm)
            channel 1 is Chappuis band (27%, 450 - 800 nm)
            channel 2 is remaining radiation (72%)
        '''
        super(ThreeBandSW, self).__init__(emissivity_sfc=emissivity_sfc, **kwargs)
        #   fraction of the total solar flux in each band:
        self.band_fraction = np.array([0.01, 0.27, 0.72])
        if 'CO2' not in self.absorber_vmr:
            self.absorber_vmr['CO2'] = 380.E-6 * np.ones_like(self.Tatm)
        if 'O3' not in self.absorber_vmr:
            self.absorber_vmr['O3'] = np.zeros_like(self.Tatm)
        if 'H2O' not in self.absorber_vmr:
            self.absorber_vmr['H2O'] = self.q
        ##  absorption cross-sections in m**2 / kg
        O3 = np.array([200.E-24, 0.285E-24, 0.]) * const.Rd / const.kBoltzmann
        #self.absorption_cross_section['O3'] = np.reshape(O3,
        #    (self.num_channels, 1))
        #H2O = np.array([0.002, 0.002, 0.002])
        H2O = np.array([0., 0., 0.001])
        for n in range(self.Tatm.domain.numdims):
            H2O = H2O[:, np.newaxis]
            O3 = O3[:, np.newaxis]
        self.absorption_cross_section['O3'] = O3
        self.absorption_cross_section['H2O'] = H2O
        #self.absorption_cross_section['H2O'] = np.reshape(H2O,
        #    (self.num_channels, 1))
        self.absorption_cross_section['CO2'] = \
            np.zeros_like(self.absorption_cross_section['O3'])
        self.cosZen = 0.5  # cosine of the average solar zenith angle

    @property
    def emissivity(self):
        # This ensures that emissivity is always zero for shortwave classes
        return np.zeros_like(self.absorptivity)


class FourBandSW(NbandRadiation):
    #  The most recent RMCM program uses a four-channel SW model
    #  this is probably a better way to do it... distinguishes between
    #  visible band with no absorption and near-infrared with weak H2O absorption
    #   But this needs some tuning and better documentation
    def __init__(self, emissivity_sfc=0., **kwargs):
        '''A four-band mdoel for shortwave radiation.

        The spectral decomposition used here is largely based on the
        "Moist Radiative-Convective Model" by Aarnout van Delden, Utrecht University
        a.j.vandelden@uu.nl
        http://www.staff.science.uu.nl/~delde102/RCM.htm

        Four SW channels:
            channel 0 is Hartley and Huggins band (UV, 6%, <340 nm)
            channel 1 is part of visible with no O3 absorption (14%, 340 - 500 nm)
            channel 2 is Chappuis band (27%, 500 - 700 nm)
            channel 3 is near-infrared (53%, > 700 nm)
        '''
        super(FourBandSW, self).__init__(emissivity_sfc=emissivity_sfc, **kwargs)
        #   fraction of the total solar flux in each band:
        self.band_fraction = np.array([0.06, 0.14, 0.27, 0.53])
        if 'CO2' not in self.absorber_vmr:
            self.absorber_vmr['CO2'] = 380.E-6 * np.ones_like(self.Tatm)
        if 'O3' not in self.absorber_vmr:
            self.absorber_vmr['O3'] = np.zeros_like(self.Tatm)
        if 'H2O' not in self.absorber_vmr:
            self.absorber_vmr['H2O'] = self.q
        ##  absorption cross-sections in m**2 / kg
        O3 = np.array([200.E-24, 0., 0.285E-24, 0.]) * const.Rd / const.kBoltzmann
        #self.absorption_cross_section['O3'] = np.reshape(O3,
        #    (self.num_channels, 1))
        H2O = np.array([0., 0., 0., 0.0012])
        for n in range(self.Tatm.domain.numdims):
            H2O = H2O[:, np.newaxis]
            O3 = O3[:, np.newaxis]
        self.absorption_cross_section['O3'] = O3
        self.absorption_cross_section['H2O'] = H2O
        #self.absorption_cross_section['H2O'] = np.reshape(H2O,
        #    (self.num_channels, 1))
        self.absorption_cross_section['CO2'] = \
            np.zeros_like(self.absorption_cross_section['O3'])
        self.cosZen = 0.5  # cosine of the average solar zenith angle

    @property
    def emissivity(self):
        # This ensures that emissivity is always zero for shortwave classes
        return np.zeros_like(self.absorptivity)



class FourBandLW(NbandRadiation):
    def __init__(self, **kwargs):
        '''Closely following SPEEDY / MITgcm longwave model
        band 0 is window region (between 8.5 and 11 microns)
        band 1 is CO2 channel (the band of strong absorption by CO2 around 15 microns)
        band 2 is weak H2O channel (aggregation of spectral regions with weak to moderate absorption by H2O)
        band 3 is strong H2O channel (aggregation of regions with strong absorption by H2O)
        '''
        super(FourBandLW, self).__init__(**kwargs)
        #  SPEEDY uses an approximation to the Planck function
        #  and the band fraction for every emission is calculated from
        #  its current temperature
        #  Here for simoplicity we'll just set an average band_fraction
        #  and hold it fixed
        Tarray = np.linspace(-30, 30) + 273.15
        self.band_fraction = np.mean(SPEEDY_band_fraction(Tarray), axis=1)

        # defaults from MITgcm/aim:
        # these are layer absorptivities per dp = 10^5 Pa
        # the water vapor terms are expressed for dq = 1 g/kg
        ABLWIN = 0.7
        ABLCO2 = 4.0
        ABLWV1 = 0.7
        ABLWV2 = 50.0
        #  I'm going to assume that the absorption in window region is by O3.
        #   not sure if this makes any sense... maybe it should be zero
        O3 = np.array([ABLWIN, 0., 0., 0.]) / 1E5 * const.g / 5E-6
        self.absorption_cross_section['O3'] = np.reshape(O3,
            (self.num_channels, 1))
        # the CO2 mixing ratio for which SPEEDY / MITgcm is tuned...
        #   not clear what this number should be
        AIMCO2 = 380E-6
        CO2 = np.array([0., ABLCO2, 0., 0.]) / 1E5 * const.g / AIMCO2
        #self.absorption_cross_section['CO2'] = np.reshape(CO2,
        #    (self.num_channels, 1))
        # Need to multiply by 1E3 for H2O fields because we use kg/kg for mixing ratio
        H2O = np.array([0., 0., ABLWV1, ABLWV2]) / 1E5 * const.g * 1E3
        #self.absorption_cross_section['H2O'] = np.reshape(H2O,
        #    (self.num_channels, 1))
        for n in range(self.Tatm.domain.numdims):
            CO2 = CO2[:, np.newaxis]
            O3 = O3[:, np.newaxis]
            H2O = H2O[:, np.newaxis]
        self.absorption_cross_section.update({'CO2': CO2, 'H2O': H2O, 'O3': O3})
        if 'CO2' not in self.absorber_vmr:
            self.absorber_vmr['CO2'] = 380.E-6 * np.ones_like(self.Tatm)
        if 'O3' not in self.absorber_vmr:
            self.absorber_vmr['O3'] = np.zeros_like(self.Tatm)
        if 'H2O' not in self.absorber_vmr:
            self.absorber_vmr['H2O'] = self.q


def SPEEDY_band_fraction(T):
    '''Python / numpy implementation of the formula used by SPEEDY and MITgcm
    to partition longwave emissions into 4 spectral bands.

    Input: temperature in Kelvin

    returns: a four-element array of band fraction

    Reproducing here the FORTRAN code from MITgcm/pkg/aim_v23/phy_radiat.F

    .. code-block:: fortran

    	      EPS3=0.95 _d 0

    	      DO JTEMP=200,320
    	        FBAND(JTEMP,0)= EPSLW
    	        FBAND(JTEMP,2)= 0.148 _d 0 - 3.0 _d -6 *(JTEMP-247)**2
    	        FBAND(JTEMP,3)=(0.375 _d 0 - 5.5 _d -6 *(JTEMP-282)**2)*EPS3
    	        FBAND(JTEMP,4)= 0.314 _d 0 + 1.0 _d -5 *(JTEMP-315)**2
    	        FBAND(JTEMP,1)= 1. _d 0 -(FBAND(JTEMP,0)+FBAND(JTEMP,2)
    	     &                           +FBAND(JTEMP,3)+FBAND(JTEMP,4))
    	      ENDDO

    	      DO JB=0,NBAND
    	        DO JTEMP=lwTemp1,199
    	          FBAND(JTEMP,JB)=FBAND(200,JB)
    	        ENDDO
    	        DO JTEMP=321,lwTemp2
    	          FBAND(JTEMP,JB)=FBAND(320,JB)
    	        ENDDO
    	      ENDDO

    '''
    # EPSLW is the fraction of longwave emission that goes directly to space
    #  It is set to zero by default in MITgcm code. We won't use it here.
    Tarray = np.array(T)
    Tarray = np.minimum(Tarray, 320.)
    Tarray = np.maximum(Tarray, 200.)
    num_band = 4
    dims = [num_band]
    dims.extend(Tarray.shape)
    FBAND = np.zeros(dims)

    EPS2=0.95
    FBAND[1,:] = 0.148 - 3.0E-6 *(T-247.)**2
    FBAND[2,:] = (0.375 - 5.5E-6 *(T-282.)**2)*EPS2
    FBAND[3,:] = 0.314 + 1.0E-5 *(T-315.)**2
    FBAND[0,:] = 1. - np.sum(FBAND, axis=0)
    return FBAND



from __future__ import division, print_function, absolute_import
import numpy as np
import warnings
from climlab import constants as const
from climlab.radiation.radiation import _Radiation_SW
from .utils import _prepare_general_arguments
from .utils import _climlab_to_rrtm, _climlab_to_rrtm_sfc, _rrtm_to_climlab
# These values will get overridden by reading from Fortran extension
nbndsw = 1; naerec = 1; ngptsw = 1;
try:
    #  The compiled fortran extension module
    from climlab_rrtmg import rrtmg_sw as _rrtmg_sw
    nbndsw = int(_rrtmg_sw.parrrsw.nbndsw)
    naerec = int(_rrtmg_sw.parrrsw.naerec)
    ngptsw = int(_rrtmg_sw.parrrsw.ngptsw)
    #  Initialize absorption data
    _rrtmg_sw.climlab_rrtmg_sw_ini(const.cp)
except:
    warnings.warn('Cannot import and initialize compiled Fortran extension, RRTMG_SW module will not be functional.')


class RRTMG_SW(_Radiation_SW):
    def __init__(self,
            # GENERAL, used in both SW and LW
            icld = 1,    # Cloud overlap method, 0: Clear only, 1: Random, 2,  Maximum/random] 3: Maximum
            irng = 1,  # more monte carlo stuff
            permuteseed = 150,
            dyofyr = 0,       # day of the year used to get Earth/Sun distance (if not adjes)
            inflgsw  = 2,
            iceflgsw = 1,
            liqflgsw = 1,
            tauc = 0.,
            ssac = 0.,  # In-cloud single scattering albedo
            asmc = 0.,  # In-cloud asymmetry parameter
            fsfc = 0.,  # In-cloud forward scattering fraction (delta function pointing forward "forward peaked scattering")
            # AEROSOLS
            iaer = 0,   #! Aerosol option flag
                        #!    0: No aerosol
                        #!    6: ECMWF method: use six ECMWF aerosol types input aerosol optical depth at 0.55 microns for each aerosol type (ecaer)
                        #!    10:Input aerosol optical properties: input total aerosol optical depth, single scattering albedo and asymmetry parameter (tauaer, ssaaer, asmaer) directly
            tauaer = 0.,   # Aerosol optical depth (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
            ssaaer = 0.,   # Aerosol single scattering albedo (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
            asmaer = 0.,   # Aerosol asymmetry parameter (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
            ecaer  = 0.,   # Aerosol optical depth at 0.55 micron (iaer=6 only), Dimensions,  (ncol,nlay,naerec)] #  (non-delta scaled)
            # new arguments for RRTMG_SW version 4.0
            isolvar = -1,    # ! Flag for solar variability method
                            # !   -1 = (when scon .eq. 0.0): No solar variability
                            # !        and no solar cycle (Kurucz solar irradiance
                            # !        of 1368.22 Wm-2 only);
                            # !        (when scon .ne. 0.0): Kurucz solar irradiance
                            # !        scaled to scon and solar variability defined
                            # !        (optional) by setting non-zero scale factors
                            # !        for each band in bndsolvar
                            # !    0 = (when SCON .eq. 0.0): No solar variability
                            # !        and no solar cycle (NRLSSI2 solar constant of
                            # !        1360.85 Wm-2 for the 100-50000 cm-1 spectral
                            # !        range only), with facular and sunspot effects
                            # !        fixed to the mean of Solar Cycles 13-24;
                            # !        (when SCON .ne. 0.0): No solar variability
                            # !        and no solar cycle (NRLSSI2 solar constant of
                            # !        1360.85 Wm-2 for the 100-50000 cm-1 spectral
                            # !        range only), is scaled to SCON
                            # !    1 = Solar variability (using NRLSSI2  solar
                            # !        model) with solar cycle contribution
                            # !        determined by fraction of solar cycle
                            # !        with facular and sunspot variations
                            # !        fixed to their mean variations over the
                            # !        average of Solar Cycles 13-24;
                            # !        two amplitude scale factors allow
                            # !        facular and sunspot adjustments from
                            # !        mean solar cycle as defined by indsolvar
                            # !    2 = Solar variability (using NRLSSI2 solar
                            # !        model) over solar cycle determined by
                            # !        direct specification of Mg (facular)
                            # !        and SB (sunspot) indices provided
                            # !        in indsolvar (scon = 0.0 only)
                            # !    3 = (when scon .eq. 0.0): No solar variability
                            # !        and no solar cycle (NRLSSI2 solar irradiance
                            # !        of 1360.85 Wm-2 only);
                            # !        (when scon .ne. 0.0): NRLSSI2 solar irradiance
                            # !        scaled to scon and solar variability defined
                            # !        (optional) by setting non-zero scale factors
                            # !        for each band in bndsolvar
            indsolvar = np.ones(2),      # Facular and sunspot amplitude scale factors (isolvar=1),
                                         # or Mg and SB indices (isolvar=2)
            bndsolvar = np.ones(nbndsw), # Solar variability scale factors for each shortwave band
            solcycfrac = 1.,              # Fraction of averaged solar cycle (0-1) at current time (isolvar=1)
            **kwargs):
        super(RRTMG_SW, self).__init__(**kwargs)
        #  define INPUTS
        self.add_input('icld', icld)
        self.add_input('irng', irng)
        self.add_input('permuteseed', permuteseed)
        self.add_input('dyofyr', dyofyr)
        self.add_input('inflgsw', inflgsw)
        self.add_input('iceflgsw', iceflgsw)
        self.add_input('liqflgsw', liqflgsw)
        self.add_input('iaer', iaer)
        self.add_input('tauc', tauc)
        self.add_input('ssac', ssac)
        self.add_input('asmc', asmc)
        self.add_input('fsfc', fsfc)
        self.add_input('tauaer', tauaer)
        self.add_input('ssaaer', ssaaer)
        self.add_input('asmaer', asmaer)
        self.add_input('ecaer', ecaer)
        self.add_input('isolvar', isolvar)
        self.add_input('indsolvar', indsolvar)
        self.add_input('bndsolvar', bndsolvar)
        self.add_input('solcycfrac', solcycfrac)

    def _prepare_sw_arguments(self):
        #  prepare insolation
        #  CLIMLAB provides local insolation, solar constant and
        #  (optionally, maybe, not yet implemented) the adjustment factor for Sun-Earth distance
        #  which is (dbar / d)^2 in Hartmann's notation
        #   the factor by which the total irradiance differs from mean annual solar constant
        #   due to time of year and elliptical orbit
        #
        #  RRTMG_SW expects solar constant, adjustment factor, and cosine of zenith angle
        #
        #  actually for now let's just let all these things be specified as inputs

        #  scalar integer arguments
        icld = self.icld
        irng = self.irng
        permuteseed = self.permuteseed
        inflgsw = self.inflgsw
        iceflgsw = self.iceflgsw
        liqflgsw = self.liqflgsw
        dyofyr = self.dyofyr
        isolvar = self.isolvar
        solcycfrac = self.solcycfrac
        iaer = self.iaer
        #  scalar real arguments
        adjes = self.eccentricity_factor
        scon = self.S0
        indsolvar = self.indsolvar
        bndsolvar = self.bndsolvar

        (ncol, nlay, play, plev, tlay, tlev, tsfc,
        h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr, cfc11vmr,
        cfc12vmr, cfc12vmr, cfc22vmr, ccl4vmr,
        cldfrac, ciwp, clwp, relq, reic) = _prepare_general_arguments(self)

        aldif = _climlab_to_rrtm_sfc(self.aldif, self.Ts)
        aldir = _climlab_to_rrtm_sfc(self.aldir, self.Ts)
        asdif = _climlab_to_rrtm_sfc(self.asdif, self.Ts)
        asdir = _climlab_to_rrtm_sfc(self.asdir, self.Ts)
        coszen = _climlab_to_rrtm_sfc(self.coszen, self.Ts)
        #  These arrays have an extra dimension for number of bands
        # in-cloud optical depth [nbndsw,ncol,nlay]
        tauc = _climlab_to_rrtm(self.tauc * np.ones_like(self.Tatm))
        #  broadcast to get [nbndsw,ncol,nlay]
        tauc = tauc * np.ones([nbndsw,ncol,nlay])
        # In-cloud single scattering albedo, same operation
        ssac = _climlab_to_rrtm(self.ssac * np.ones_like(self.Tatm)) * np.ones([nbndsw,ncol,nlay])
        # In-cloud asymmetry parameter
        asmc = _climlab_to_rrtm(self.asmc * np.ones_like(self.Tatm)) * np.ones([nbndsw,ncol,nlay])
        # In-cloud forward scattering fraction (delta function pointing forward "forward peaked scattering")
        fsfc = _climlab_to_rrtm(self.fsfc * np.ones_like(self.Tatm)) * np.ones([nbndsw,ncol,nlay])
        # Aerosol optical depth (iaer=10 only), (ncol,nlay,nbndsw)] #  (non-delta scaled)
        tauaer = _climlab_to_rrtm(self.tauaer * np.ones_like(self.Tatm))
        #  broadcast and transpose to get [ncol,nlay,nbndsw]
        tauaer = np.transpose(tauaer * np.ones([nbndsw,ncol,nlay]), (1,2,0))
        # Aerosol single scattering albedo (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
        ssaaer = np.transpose(_climlab_to_rrtm(self.ssaaer * np.ones_like(self.Tatm)) * np.ones([nbndsw,ncol,nlay]), (1,2,0))
        # Aerosol asymmetry parameter (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
        asmaer = np.transpose(_climlab_to_rrtm(self.asmaer * np.ones_like(self.Tatm)) * np.ones([nbndsw,ncol,nlay]), (1,2,0))
        # Aerosol optical depth at 0.55 micron (iaer=6 only), Dimensions,  (ncol,nlay,naerec)] #  (non-delta scaled)
        ecaer = np.transpose(_climlab_to_rrtm(self.ecaer * np.ones_like(self.Tatm)) * np.ones([naerec,ncol,nlay]), (1,2,0))

        args = [ncol, nlay, icld, iaer, permuteseed, irng,
                play, plev, tlay, tlev, tsfc,
                h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr,
                aldif, aldir, asdif, asdir, coszen, adjes, dyofyr, scon, isolvar,
                indsolvar, bndsolvar, solcycfrac,
                inflgsw, iceflgsw, liqflgsw,
                cldfrac, ciwp, clwp, reic, relq, tauc, ssac, asmc, fsfc,
                tauaer, ssaaer, asmaer, ecaer,]

        return args

    def _compute_heating_rates(self):
        '''Prepare arguments and call the RRTGM_SW driver to calculate
        radiative fluxes and heating rates'''
        (ncol, nlay, icld, iaer, permuteseed, irng,
         play, plev, tlay, tlev, tsfc,
         h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr,
         aldif, aldir, asdif, asdir, coszen, adjes, dyofyr, scon, isolvar,
         indsolvar, bndsolvar, solcycfrac,
         inflgsw, iceflgsw, liqflgsw,
         cldfrac, ciwp, clwp, reic, relq, tauc, ssac, asmc, fsfc,
         tauaer, ssaaer, asmaer, ecaer,) = self._prepare_sw_arguments()
        if icld == 0:  # clear-sky only
            cldfmcl = np.zeros((ngptsw,ncol,nlay))
            ciwpmcl = np.zeros((ngptsw,ncol,nlay))
            clwpmcl = np.zeros((ngptsw,ncol,nlay))
            reicmcl = np.zeros((ncol,nlay))
            relqmcl = np.zeros((ncol,nlay))
            taucmcl = np.zeros((ngptsw,ncol,nlay))
            ssacmcl = np.zeros((ngptsw,ncol,nlay))
            asmcmcl = np.zeros((ngptsw,ncol,nlay))
            fsfcmcl = np.zeros((ngptsw,ncol,nlay))
        else:
            #  Call the Monte Carlo Independent Column Approximation (McICA, Pincus et al., JC, 2003)
            (cldfmcl, ciwpmcl, clwpmcl, reicmcl, relqmcl, taucmcl,
            ssacmcl, asmcmcl, fsfcmcl) = _rrtmg_sw.climlab_mcica_subcol_sw(
                            ncol, nlay, icld, permuteseed, irng, play,
                            cldfrac, ciwp, clwp, reic, relq, tauc, ssac, asmc, fsfc)
        #  Call the RRTMG_SW driver to compute radiative fluxes
        (swuflx, swdflx, swhr, swuflxc, swdflxc, swhrc) = \
            _rrtmg_sw.climlab_rrtmg_sw(ncol, nlay, icld, iaer,
                play, plev, tlay, tlev, tsfc,
                h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr,
                asdir, asdif, aldir, aldif,
                coszen, adjes, dyofyr, scon, isolvar,
                inflgsw, iceflgsw, liqflgsw, cldfmcl,
                taucmcl, ssacmcl, asmcmcl, fsfcmcl,
                ciwpmcl, clwpmcl, reicmcl, relqmcl,
                tauaer, ssaaer, asmaer, ecaer,
                bndsolvar, indsolvar, solcycfrac)
        #  Output is all (ncol,nlay+1) or (ncol,nlay)
        self.SW_flux_up = _rrtm_to_climlab(swuflx) + 0.*self.SW_flux_up
        self.SW_flux_down = _rrtm_to_climlab(swdflx) + 0.*self.SW_flux_down
        self.SW_flux_up_clr = _rrtm_to_climlab(swuflxc) + 0.*self.SW_flux_up_clr
        self.SW_flux_down_clr = _rrtm_to_climlab(swdflxc) + 0.*self.SW_flux_down_clr
        #  Compute quantities derived from fluxes, including ASR
        self._compute_SW_flux_diagnostics()
        #  calculate heating rates from flux divergence
        SWheating_Wm2 = np.array(-np.diff(self.SW_flux_net, axis=-1)) + 0.*self.Tatm
        SWheating_clr_Wm2 = np.array(-np.diff(self.SW_flux_net_clr, axis=-1)) + 0.*self.Tatm
        self.heating_rate['Ts'] = np.array(self.SW_flux_net[..., -1, np.newaxis]) + 0.*self.Ts
        self.heating_rate['Tatm'] = SWheating_Wm2
        #  Convert to K / day
        Catm = self.Tatm.domain.heat_capacity
        self.TdotSW = SWheating_Wm2 / Catm * const.seconds_per_day
        self.TdotSW_clr = SWheating_clr_Wm2 / Catm * const.seconds_per_day



from __future__ import division, print_function, absolute_import
import numpy as np
import warnings
from climlab import constants as const
from climlab.radiation.radiation import _Radiation_LW
from climlab.domain import Field, Axis, domain
from .utils import _prepare_general_arguments
from .utils import _climlab_to_rrtm, _rrtm_to_climlab
# These values will get overridden by reading from Fortran extension
nbndlw = 1; ngptlw = 1;
try:
    #  The compiled fortran extension module
    from climlab_rrtmg import rrtmg_lw as _rrtmg_lw
    nbndlw = int(_rrtmg_lw.parrrtm.nbndlw)
    ngptlw = int(_rrtmg_lw.parrrtm.ngptlw)
    #  Initialize absorption data
    _rrtmg_lw.climlab_rrtmg_lw_ini(const.cp)
except:
    warnings.warn('Cannot import and initialize compiled Fortran extension, RRTMG_LW module will not be functional.')

# Longwave spectral band limits (wavenumbers in cm^-1)
# Data copied from rrtmg_lw_v4.85/gcm_model/src/rrtmg_lw_init.f90
wavenum_bounds = np.array([  10., 350., 500., 630., 700., 820.,
                      980.,1080.,1180.,1390.,1480.,1800.,
                     2080.,2250.,2380.,2600.,3250.])
wavenum_delta = np.diff(wavenum_bounds)


class RRTMG_LW(_Radiation_LW):
    def __init__(self,
            # GENERAL, used in both SW and LW
            icld = 1,    # Cloud overlap method, 0: Clear only, 1: Random, 2,  Maximum/random] 3: Maximum
            irng = 1,  # more monte carlo stuff
            idrv = 0,  # whether to also calculate the derivative of flux with respect to surface temp
            permuteseed =  300,
            inflglw  = 2,
            iceflglw = 1,
            liqflglw = 1,
            tauc = 0.,  # in-cloud optical depth
            tauaer = 0.,   # Aerosol optical depth at mid-point of LW spectral bands
            return_spectral_olr = False, # Whether or not to return OLR averaged over each band
            **kwargs):
        super(RRTMG_LW, self).__init__(**kwargs)
        #  define INPUTS
        self.add_input('icld', icld)
        self.add_input('irng', irng)
        self.add_input('idrv', idrv)
        self.add_input('permuteseed', permuteseed)
        self.add_input('inflglw', inflglw)
        self.add_input('iceflglw', iceflglw)
        self.add_input('liqflglw', liqflglw)
        self.add_input('tauc', tauc)
        self.add_input('tauaer', tauaer)
        self.add_input('return_spectral_olr', return_spectral_olr)

        # Spectrally-decomposed OLR
        if self.return_spectral_olr:
            # Adjust output flag
            self._ispec = 1  # Spectral OLR output flag, 0: only calculate total fluxes, 1: also return spectral OLR
            # set up appropriately sized Field object to store the spectral OLR diagnostic
            wavenum_ax = Axis(axis_type='abstract', bounds=wavenum_bounds)
            spectral_axes = {**self.OLR.domain.axes, 'wavenumber': wavenum_ax}
            spectral_domain = domain._Domain(axes=spectral_axes)
            #  HACK need to reorder axes in the domain object
            shape = list(self.OLR.shape)
            shape.append(wavenum_ax.num_points)
            spectral_domain.shape = tuple(shape)
            spectral_domain.axis_index = {**self.OLR.domain.axis_index, 'wavenumber': len(shape)-1}
            # This ensures that the spectral dimension (length: nbndlw) is appended after existing grid dimensions
            blank_field = Field((self.OLR[...,np.newaxis] * wavenum_delta), domain=spectral_domain)
            self.add_diagnostic('OLR_spectral', blank_field)
        else:
            self._ispec = 0,  # Spectral OLR output flag, 0: only calculate total fluxes, 1: also return spectral OLR


    def _prepare_lw_arguments(self):
        #  scalar integer arguments
        icld  = self.icld
        ispec = self._ispec
        irng  = self.irng
        idrv  = self.idrv
        permuteseed = self.permuteseed
        inflglw = self.inflglw
        iceflglw = self.iceflglw
        liqflglw = self.liqflglw

        (ncol, nlay, play, plev, tlay, tlev, tsfc,
        h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr, cfc11vmr,
        cfc12vmr, cfc12vmr, cfc22vmr, ccl4vmr,
        cldfrac, ciwp, clwp, relq, reic) = _prepare_general_arguments(self)
        # surface emissivity
        emis = self.emissivity * np.ones((ncol,nbndlw))
        #  These arrays have an extra dimension for number of bands
        # in-cloud optical depth [nbndlw,ncol,nlay]
        tauc = _climlab_to_rrtm(self.tauc * np.ones_like(self.Tatm))
        #  broadcast to get [nbndlw,ncol,nlay]
        tauc = tauc * np.ones([nbndlw,ncol,nlay])
        # Aerosol optical depth at mid-point of LW spectral bands [ncol,nlay,nbndlw]
        tauaer = _climlab_to_rrtm(self.tauaer * np.ones_like(self.Tatm))
        #  broadcast and transpose to get [ncol,nlay,nbndlw]
        tauaer = np.transpose(tauaer * np.ones([nbndlw,ncol,nlay]), (1,2,0))
        args = [ncol, nlay, icld, ispec, permuteseed, irng, idrv, const.cp,
                play, plev, tlay, tlev, tsfc,
                h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr,
                cfc11vmr, cfc12vmr, cfc22vmr, ccl4vmr, emis,
                inflglw, iceflglw, liqflglw,
                cldfrac, ciwp, clwp, reic, relq, tauc, tauaer,]
        return args

    def _compute_heating_rates(self):
        '''Prepare arguments and call the RRTGM_LW driver to calculate
        radiative fluxes and heating rates'''
        (ncol, nlay, icld, ispec, permuteseed, irng, idrv, cp,
                play, plev, tlay, tlev, tsfc,
                h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr,
                cfc11vmr, cfc12vmr, cfc22vmr, ccl4vmr, emis,
                inflglw, iceflglw, liqflglw,
                cldfrac, ciwp, clwp, reic, relq, tauc, tauaer,) = self._prepare_lw_arguments()
        if icld == 0:  # clear-sky only
            cldfmcl = np.zeros((ngptlw,ncol,nlay))
            ciwpmcl = np.zeros((ngptlw,ncol,nlay))
            clwpmcl = np.zeros((ngptlw,ncol,nlay))
            reicmcl = np.zeros((ncol,nlay))
            relqmcl = np.zeros((ncol,nlay))
            taucmcl = np.zeros((ngptlw,ncol,nlay))
        else:
            #  Call the Monte Carlo Independent Column Approximation (McICA, Pincus et al., JC, 2003)
            (cldfmcl, ciwpmcl, clwpmcl, reicmcl, relqmcl, taucmcl) = \
                _rrtmg_lw.climlab_mcica_subcol_lw(
                            ncol, nlay, icld,
                            permuteseed, irng, play,
                            cldfrac, ciwp, clwp, reic, relq, tauc)
            #  Call the RRTMG_LW driver to compute radiative fluxes
        (olr_sr, uflx, dflx, hr, uflxc, dflxc, hrc, duflx_dt, duflxc_dt) = \
            _rrtmg_lw.climlab_rrtmg_lw(ncol, nlay, icld, ispec, idrv,
                 play, plev, tlay, tlev, tsfc,
                 h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr,
                 cfc11vmr, cfc12vmr, cfc22vmr, ccl4vmr, emis,
                 inflglw, iceflglw, liqflglw, cldfmcl,
                 taucmcl, ciwpmcl, clwpmcl, reicmcl, relqmcl,
                 tauaer)
        #  Output is all (ncol,nlay+1) or (ncol,nlay)
        self.LW_flux_up = _rrtm_to_climlab(uflx) + 0.*self.LW_flux_up
        self.LW_flux_down = _rrtm_to_climlab(dflx) + 0.*self.LW_flux_down
        self.LW_flux_up_clr = _rrtm_to_climlab(uflxc) + 0.*self.LW_flux_up_clr
        self.LW_flux_down_clr = _rrtm_to_climlab(dflxc) + 0.*self.LW_flux_down_clr
        #  Compute quantities derived from fluxes, including OLR
        self._compute_LW_flux_diagnostics()
        # Except for spectrally-decomposed TOA flux, olr_sr (ncol, nbndlw)
        if self.return_spectral_olr:
            #  Need to deal with broadcasting for two different cases: single column and latitude axis
            # case single column: self.OLR is (1,),  self.OLR_spectral is (1, nbndlw),  olr_sr is (1,nbndlw)
            #  squeeze olr_sr down to (nbndlw,)
            # then use np.squeeze(olr_sr)[..., np.newaxis, :] to get back to (1, nbndlw)
            # case latitude axis: self.OLR is (num_lat,1), self.OLR_spectral is (num_lat, 1, nbndlw), olr_sr is (num_lat, nbndlw)
            #  np.squeeze(olr_sr) has no effect in this case
            # add the newaxis because the domain has a size-1 depth axis ---> (num_lat, 1, nbndlw)
            self.OLR_spectral = np.squeeze(olr_sr)[...,np.newaxis,:] + 0.*self.OLR_spectral
        #  calculate heating rates from flux divergence
        LWheating_Wm2 = np.array(np.diff(self.LW_flux_net, axis=-1)) + 0.*self.Tatm
        LWheating_clr_Wm2 = np.array(np.diff(self.LW_flux_net_clr, axis=-1)) + 0.*self.Tatm
        self.heating_rate['Ts'] = np.array(-self.LW_flux_net[..., -1, np.newaxis]) + 0.*self.Ts
        self.heating_rate['Tatm'] = LWheating_Wm2
        #  Convert to K / day
        Catm = self.Tatm.domain.heat_capacity
        self.TdotLW = LWheating_Wm2 / Catm * const.seconds_per_day
        self.TdotLW_clr = LWheating_clr_Wm2 / Catm * const.seconds_per_day



"""
The object ``climlab.solar.orbital.OrbitalTable`` is an ``xarray.Dataset``
holding orbital data (**eccentricity**, **obliquity**, and **longitude of perihelion**)
for the past 5 Myears. The data are from :cite:`Berger_1991`.

Data are read from the file ``orbit91``, which was originally obtained from
<https://www1.ncdc.noaa.gov/pub/data/paleo/climate_forcing/orbital_variations/insolation/>
If the file isn't found locally, the module will attempt to read it remotely
from the above URL.

A subclass ``climlab.solar.orbital.long.OrbitalTable``
works with La2004 orbital data for
-51 to +21 Myears as calculated by :cite:`Laskar_2004`.
See <http://vo.imcce.fr/insola/earth/online/earth/La2004/README.TXT>

(Breaking change from climlab 0.7.0 and previous)

    :Example:
        Load orbital data from the past 5 Myears::

            # Load the data
            from climlab.solar.orbital import OrbitalTable

            # Examine the xarray object
            print(OrbitalTable)

            # Get a timeseries of obliquity
            print(OrbitalTable.obliquity)

            # Get the orbital data for a specific year, 10 kyear before present:
            print(OrbitalTable.interp(kyear=-10))

            # Get the long orbital table data
            from climlab.solar.orbital.long import OrbitalTable as LongTable
            print(LongTable)
"""
from __future__ import division, print_function, absolute_import
import numpy as np
import pandas as pd
import xarray as xr
from .table import _get_Berger_data

OrbitalTable = _get_Berger_data()



'''
Modules for atmospheric convection.

For simple adjustment of temperature to a prescribed lapse rate, use :class:`~climlab.convection.ConvectiveAdjustment`

For a full convection scheme including interactive water vapor, use :class:`~climlab.convection.EmanuelConvection`
'''
from __future__ import absolute_import
from .convadj import ConvectiveAdjustment
from .emanuel_convection import EmanuelConvection



"""This module contains general-purpose routines for computing
daily-average incoming solar radiation at the top of the atmosphere.

    :Example:
        Compute the timeseries of insolation at 65N at summer
        solstice over the past 5 Myears::

            import numpy as np
            from climlab.solar.orbital import OrbitalTable
            from climlab.solar.insolation import daily_insolation

            # array with specified kyears (can be plain numpy or xarray.DataArray)
            years = np.linspace(-5000, 0, 5001)

            # subset of orbital parameters for specified time
            orb = OrbitalTable.interp(kyear=years)

            # insolation values for past 5 Myears at 65N at summer solstice (day 172)
            S65 = daily_insolation(lat=65, day=172, orb=orb)
            # returns an xarray.DataArray object with insolation values in W/m2

.. note::

    Ported and modified from MATLAB code daily_insolation.m         \n
    *Original authors:*                                             \n

     Ian Eisenman and Peter Huybers, Harvard University, August 2006

    Available online at http://eisenman.ucsd.edu/code/daily_insolation.m

If using calendar days, solar longitude is found using an
approximate solution to the differential equation representing conservation
of angular momentum (Kepler's Second Law).  Given the orbital parameters
and solar longitude, daily average insolation is calculated exactly
following :cite:`Berger_1978`. Further references: :cite:`Berger_1991`.

"""

from __future__ import division
import numpy as np
from climlab import constants as const
from numpy import sqrt, deg2rad, rad2deg, sin, cos, tan, arcsin, arccos, pi
import xarray as xr


def daily_insolation(lat, day, orb=const.orb_present, S0=const.S0, day_type=1):
    """Compute daily average insolation given latitude, time of year and orbital parameters.

    Orbital parameters can be interpolated to any time in the last 5 Myears with
    ``climlab.solar.orbital.OrbitalTable`` (see example above).

    Longer orbital tables are available with ``climlab.solar.orbital.LongOrbitalTable``

    Inputs can be scalar, ``numpy.ndarray``, or ``xarray.DataArray``.

    The return value will be ``numpy.ndarray`` if **all** the inputs are ``numpy``.
    Otherwise ``xarray.DataArray``.

    **Function-call argument** \n

    :param array lat:       Latitude in degrees (-90 to 90).
    :param array day:       Indicator of time of year. See argument ``day_type``
                            for details about format.
    :param dict orb:        a dictionary with three members (as provided by
                            ``climlab.solar.orbital.OrbitalTable``)

                            * ``'ecc'`` - eccentricity

                                * unit: dimensionless
                                * default value: ``0.017236``

                            * ``'long_peri'`` - longitude of perihelion (precession angle)

                                * unit: degrees
                                * default value: ``281.37``

                            * ``'obliquity'`` - obliquity angle

                                * unit: degrees
                                * default value: ``23.446``

    :param float S0:        solar constant                                  \n
                            - unit: :math:`\\textrm{W}/\\textrm{m}^2`       \n
                            - default value: ``1365.2``
    :param int day_type:    Convention for specifying time of year (+/- 1,2) [optional].

                            *day_type=1* (default):
                             day input is calendar day (1-365.24), where day 1
                             is January first. The calendar is referenced to the
                             vernal equinox which always occurs at day 80.

                            *day_type=2:*
                             day input is solar longitude (0-360 degrees). Solar
                             longitude is the angle of the Earth's orbit measured from spring
                             equinox (21 March). Note that calendar days and solar longitude are
                             not linearly related because, by Kepler's Second Law, Earth's
                             angular velocity varies according to its distance from the sun.
    :raises: :exc:`ValueError`
                            if day_type is neither 1 nor 2
    :returns:               Daily average solar radiation in unit
                            :math:`\\textrm{W}/\\textrm{m}^2`.

                            Dimensions of output are ``(lat.size, day.size, ecc.size)``
    :rtype:                 array


    Code is fully vectorized to handle array input for all arguments.       \n
    Orbital arguments should all have the same sizes.
    This is automatic if computed from
    :func:`~climlab.solar.orbital.OrbitalTable.lookup_parameters`

        For more information about computation of solar insolation see the
        :ref:`Tutorial` chapter.

     """
    # Inputs can be scalar, numpy vector, or xarray.DataArray.
    #  If numpy, convert to xarray so that it will broadcast correctly
    lat_is_xarray = True
    day_is_xarray = True
    if type(lat) is np.ndarray:
        lat_is_xarray = False
        lat = xr.DataArray(lat, coords=[lat], dims=['lat'])
    if type(day) is np.ndarray:
        day_is_xarray = False
        day = xr.DataArray(day, coords=[day], dims=['day'])
    ecc = orb['ecc']
    long_peri = orb['long_peri']
    obliquity = orb['obliquity']
    # Convert precession angle and latitude to radians
    phi = deg2rad( lat )

    # lambda_long (solar longitude) is the angular distance along Earth's orbit measured from spring equinox (21 March)
    if day_type==1: # calendar days
        lambda_long = solar_longitude(day,orb)
    elif day_type==2: #solar longitude (1-360) is specified in input, no need to convert days to longitude
        lambda_long = deg2rad(day)
    else:
        raise ValueError('Invalid day_type.')

    # Compute declination angle of the sun
    delta = arcsin(sin(deg2rad(obliquity)) * sin(lambda_long))

    #  suppress warning message generated by arccos here!
    oldsettings = np.seterr(invalid='ignore')
    # Compute Ho, the hour angle at sunrise / sunset
    #  Check for no sunrise or no sunset: Berger 1978 eqn (8),(9)
    Ho = xr.where( abs(delta)-pi/2+abs(phi) < 0., # there is sunset/sunrise
              arccos(-tan(phi)*tan(delta)),
              # otherwise figure out if it's all night or all day
              xr.where(phi*delta>0., pi, 0.) )
    # this is not really the daily average cosine of the zenith angle...
    #  it's the integral from sunrise to sunset of that quantity...
    coszen = Ho*sin(phi)*sin(delta) + cos(phi)*cos(delta)*sin(Ho)
    # Compute insolation: Berger 1978 eq (10)
    Fsw = S0/pi*( (1+ecc*cos(lambda_long -deg2rad(long_peri)))**2 / (1-ecc**2)**2 * coszen)
    if not (lat_is_xarray or day_is_xarray):
        # Dimensional ordering consistent with previous numpy code
        return Fsw.transpose().values
    else:
        return Fsw


def solar_longitude( day, orb=const.orb_present, days_per_year = None ):
    """Estimates solar longitude from calendar day.

    Method is using an approximation from :cite:`Berger_1978` section 3
    (lambda = 0 at spring equinox).

    **Function-call arguments** \n

    :param array day:           Indicator of time of year.
    :param dict orb:            a dictionary with three members (as provided by
                                :class:`~climlab.solar.orbital.OrbitalTable`)

                                * ``'ecc'`` - eccentricity

                                    * unit: dimensionless
                                    * default value: ``0.017236``

                                * ``'long_peri'`` - longitude of perihelion
                                  (precession angle)

                                    * unit: degrees
                                    * default value: ``281.37``

                                * ``'obliquity'`` - obliquity angle

                                    * unit: degrees
                                    * default value: ``23.446``
    :param float days_per_year: number of days in a year (optional)
                                (default: 365.2422)
                                Reads the length of the year from
                                :mod:`~climlab.utils.constants` if available.
    :returns:                   solar longitude ``lambda_long``
                                in dimension``( day.size, ecc.size )``
    :rtype:                     array

    Works for both scalar and vector orbital parameters.



    """
    if days_per_year is None:
        days_per_year = const.days_per_year
    ecc = orb['ecc']
    long_peri_rad = deg2rad( orb['long_peri'])
    delta_lambda = (day - 80.) * 2*pi/days_per_year
    beta = sqrt(1 - ecc**2)
    lambda_long_m = -2*((ecc/2 + (ecc**3)/8 ) * (1+beta) * sin(-long_peri_rad) -
        (ecc**2)/4 * (1/2 + beta) * sin(-2*long_peri_rad) + (ecc**3)/8 *
        (1/3 + beta) * sin(-3*long_peri_rad)) + delta_lambda
    lambda_long = ( lambda_long_m + (2*ecc - (ecc**3)/4)*sin(lambda_long_m - long_peri_rad) +
        (5/4)*(ecc**2) * sin(2*(lambda_long_m - long_peri_rad)) + (13/12)*(ecc**3)
        * sin(3*(lambda_long_m - long_peri_rad)) )
    return lambda_long



"""
attrdict contains several mapping objects that allow access to their
keys as attributes.
"""
from .mapping import AttrMap
from .dictionary import AttrDict
from .default import AttrDefault


__all__ = ['AttrMap', 'AttrDict', 'AttrDefault']



import numpy as np
import climlab
from climlab.dynamics.diffusion import MeridionalDiffusion
from climlab.utils import legendre

sfc = climlab.domain.zonal_mean_surface(num_lat=90, water_depth=10.)
lat = sfc.lat.points
initial = 12. - 40. * legendre.P2(np.sin(np.deg2rad(lat)))

# make a copy of initial so that it remains unmodified
Ts = climlab.Field(np.array(initial), domain=sfc)

# thermal diffusivity in W/m**2/degC
D = 0.55

# meridional diffusivity in 1/s
K = D / sfc.heat_capacity
d = MeridionalDiffusion(state=Ts, K=K)

d.integrate_years(1.)

import matplotlib.pyplot as plt

fig = plt.figure( figsize=(6,4))
ax = fig.add_subplot(111)
ax.set_title('Example for Meridional Diffusion')
ax.set_xlabel('latitude')
ax.set_xticks([-90,-60,-30,0,30,60,90])
ax.set_ylabel('temperature ($^{\circ}$C)')
ax.plot(lat, initial, 	label='initial')
ax.plot(lat, Ts, label='Ts (1yr)')
ax.legend(loc='best')
plt.show()





"""Object-oriented code for radiative-convective models with grey-gas radiation.

Code developed by Brian Rose, University at Albany
brose@albany.edu

Note that the column models by default represent global, time averages.
Thus the insolation is a prescribed constant.

Here is an example to implement seasonal insolation at 45 degrees North

    :Example:

        .. code-block:: python

            import climlab

            #  create the column model object
            col = climlab.GreyRadiationModel()

            #  create a new latitude axis with a single point
            lat = climlab.domain.Axis(axis_type='lat', points=45.)

            #  add this new axis to the surface domain
            col.Ts.domain.axes['lat'] = lat

            #  create a new insolation process using this domain
            Q = climlab.radiation.insolation.DailyInsolation(domains=col.Ts.domain, **col.param)

            #  replace the fixed insolation subprocess in the column model
            col.add_subprocess('insolation', Q)


This model is now a single column with seasonally varying insolation
calculated for 45N.

"""
from __future__ import division
import numpy as np
from climlab import constants as const
from climlab.process.time_dependent_process import TimeDependentProcess
from climlab.domain.initial import column_state
from climlab.domain.field import Field
from climlab.radiation.insolation import FixedInsolation
from climlab.radiation.greygas import GreyGas, GreyGasSW
from climlab.convection.convadj import ConvectiveAdjustment
from climlab.radiation.nband import ThreeBandSW, FourBandLW, FourBandSW
from climlab.radiation.water_vapor import ManabeWaterVapor

class GreyRadiationModel(TimeDependentProcess):
    def __init__(self,
                 num_lev=30,
                 num_lat=1,
                 lev=None,
                 lat=None,
                 water_depth=1.0,
                 albedo_sfc=0.299,
                 timestep=1. * const.seconds_per_day,
                 Q=341.3,
                 # absorption coefficient in m**2 / kg
                 abs_coeff=1.229E-4,
                 **kwargs):
        # Check to see if an initial state is already provided
        #  If not, make one
        if 'state' not in kwargs:
            state = column_state(num_lev, num_lat, lev, lat, water_depth)
            kwargs.update({'state': state})
        super(GreyRadiationModel, self).__init__(timestep=timestep, **kwargs)
        self.param['water_depth'] = water_depth
        self.param['albedo_sfc'] = albedo_sfc
        self.param['Q'] = Q
        self.param['abs_coeff'] = abs_coeff

        sfc = self.Ts.domain
        atm = self.Tatm.domain
        # create sub-models for longwave and shortwave radiation
        dp = self.Tatm.domain.lev.delta
        absorbLW = compute_layer_absorptivity(self.param['abs_coeff'], dp)
        absorbLW = Field(np.tile(absorbLW, sfc.shape), domain=atm)
        absorbSW = np.zeros_like(absorbLW)
        longwave = GreyGas(state=self.state, absorptivity=absorbLW,
                             albedo_sfc=0)
        shortwave = GreyGasSW(state=self.state, absorptivity=absorbSW,
                                albedo_sfc=self.param['albedo_sfc'])
        # sub-model for insolation ... here we just set constant Q
        thisQ = self.param['Q']*np.ones_like(self.Ts)
        Q = FixedInsolation(S0=thisQ, domains=sfc, **self.param)
        self.add_subprocess('LW', longwave)
        self.add_subprocess('SW', shortwave)
        self.add_subprocess('insolation', Q)
        newdiags = ['OLR',
                    'LW_down_sfc',
                    'LW_up_sfc',
                    'LW_absorbed_sfc',
                    'LW_absorbed_atm',
                    'LW_emission',
                    'ASR',
                    'SW_absorbed_sfc',
                    'SW_absorbed_atm',
                    'SW_up_sfc',
                    'SW_up_TOA',
                    'SW_down_TOA',
                    'planetary_albedo']
        for name in newdiags:
            self.add_diagnostic(name)
        # This process has to handle the coupling between
        # insolation and column radiation
        self.subprocess['SW'].flux_from_space = \
            self.subprocess['insolation'].diagnostics['insolation']

    def _compute(self):
        # set diagnostics
        self.do_diagnostics()
        # no tendencies for the parent process
        tendencies = {}
        for name, var in self.state.items():
            tendencies[name] = var * 0.
        return tendencies

    def do_diagnostics(self):
        '''Set all the diagnostics from long and shortwave radiation.'''
        self.OLR = self.subprocess['LW'].flux_to_space
        self.LW_down_sfc = self.subprocess['LW'].flux_to_sfc
        self.LW_up_sfc = self.subprocess['LW'].flux_from_sfc
        self.LW_absorbed_sfc = self.LW_down_sfc - self.LW_up_sfc
        self.LW_absorbed_atm = self.subprocess['LW'].absorbed
        self.LW_emission = self.subprocess['LW'].emission
            #  contributions to OLR from surface and atm. levels
            #self.diagnostics['OLR_sfc'] = self.flux['sfc2space']
            #self.diagnostics['OLR_atm'] = self.flux['atm2space']
        self.ASR = (self.subprocess['SW'].flux_from_space -
                    self.subprocess['SW'].flux_to_space)
        #self.SW_absorbed_sfc = (self.subprocess['surface'].SW_from_atm -
        #                            self.subprocess['surface'].SW_to_atm)
        self.SW_absorbed_atm = self.subprocess['SW'].absorbed
        self.SW_down_sfc = self.subprocess['SW'].flux_to_sfc
        self.SW_up_sfc = self.subprocess['SW'].flux_from_sfc
        self.SW_absorbed_sfc = self.SW_down_sfc - self.SW_up_sfc
        self.SW_up_TOA = self.subprocess['SW'].flux_to_space
        self.SW_down_TOA = self.subprocess['SW'].flux_from_space
        self.planetary_albedo = (self.subprocess['SW'].flux_to_space /
                                 self.subprocess['SW'].flux_from_space)


class RadiativeConvectiveModel(GreyRadiationModel):
    def __init__(self,
                 # lapse rate for convective adjustment, in K / km
                 adj_lapse_rate=6.5,
                 **kwargs):
        super(RadiativeConvectiveModel, self).__init__(**kwargs)
        self.param['adj_lapse_rate'] = adj_lapse_rate
        self.add_subprocess('convective adjustment', \
            ConvectiveAdjustment(state=self.state, **self.param))


class BandRCModel(RadiativeConvectiveModel):
    def __init__(self, **kwargs):
        super(BandRCModel, self).__init__(**kwargs)
        #  Initialize specific humidity
        h2o = ManabeWaterVapor(state=self.state, **self.param)
        self.add_subprocess('H2O', h2o)
        # q is an input field for this process, which is set by subproc
        #  (though in this sense it is actually diagnostic...)
        newinput = ['q']
        self.add_input('q')
        self.q = self.subprocess['H2O'].q

        #  initialize radiatively active gas inventories
        self.absorber_vmr = {}
        self.absorber_vmr['CO2'] = 380.E-6 * np.ones_like(self.Tatm)
        self.absorber_vmr['O3'] = np.zeros_like(self.Tatm)
        # water vapor is actually specific humidity, not VMR.
        self.absorber_vmr['H2O'] = self.q

        longwave = FourBandLW(state=self.state,
                              absorber_vmr=self.absorber_vmr,
                              albedo_sfc=0.)
        shortwave = ThreeBandSW(state=self.state,
                                absorber_vmr=self.absorber_vmr,
                                emissivity_sfc=0.,
                                albedo_sfc=self.param['albedo_sfc'])
        self.add_subprocess('LW', longwave)
        self.add_subprocess('SW', shortwave)
        # This process has to handle the coupling between
        # insolation and column radiation
        self.subprocess['SW'].flux_from_space = \
            self.subprocess['insolation'].diagnostics['insolation']


def compute_layer_absorptivity(abs_coeff, dp):
    '''Compute layer absorptivity from a constant absorption coefficient.'''
    return (2. / (1 + 2. * const.g / abs_coeff /
                         (dp * const.mb_to_Pa)))




import climlab
from climlab.dynamics.budyko_transport import BudykoTransport
from climlab import domain
from climlab.domain import field
from climlab.utils.legendre import P2
import numpy as np
import matplotlib.pyplot as plt

# create domain
sfc = domain.zonal_mean_surface(num_lat = 36)

lat = sfc.lat.points
lat_rad = np.deg2rad(lat)

# define initial temperature distribution
T0 = 15.
T2 = -20.
Ts = field.Field(T0 + T2 * P2(np.sin(lat_rad)), domain=sfc)

# create BudykoTransport process
budyko_transp = BudykoTransport(state=Ts)

### Integrate & Plot ###

fig = plt.figure( figsize=(6,4))
ax = fig.add_subplot(111)

for i in np.arange(0,3,1):  
    ax.plot(lat, budyko_transp.default, label='day %s' % (i*40))
    budyko_transp.integrate_days(40.)

ax.set_title('Standalone Budyko Transport')
ax.set_xlabel('latitude')
ax.set_xticks([-90,-60,-30,0,30,60,90])
ax.set_ylabel('temperature ($^{\circ}$C)')
ax.legend(loc='best')
plt.show()



from __future__ import division
from climlab.process.time_dependent_process import TimeDependentProcess


class ImplicitProcess(TimeDependentProcess):
    """A parent class for modules that use implicit time discretization.

    During initialization following attributes are intitialized:

    :ivar time_type:        is set to ``'implicit'``
    :vartype time_type:     str

    :ivar adjustment:       the model state adjustments due to this implicit
                            subprocess
    :vartype adjustment:    dict

    """
    def __init__(self, **kwargs):
        super(ImplicitProcess, self).__init__(**kwargs)
        self.time_type = 'implicit'
        self.adjustment = {}

    def _compute(self):
        """Computes the state variable tendencies in time for implicit processes.

        To calculate the new state the :func:`_implicit_solver()` method is
        called for daughter classes. This however returns the new state of the
        variables, not just the tendencies. Therefore, the adjustment is
        calculated which is the difference between the new and the old state
        and stored in the object's attribute adjustment.

        Calculating the new model states through solving the matrix problem
        already includes the multiplication with the timestep. The derived
        adjustment is divided by the timestep to calculate the implicit
        subprocess tendencies, which can be handeled by the
        :func:`~climlab.process.time_dependent_process.TimeDependentProcess.compute`
        method of the parent
        :class:`~climlab.process.time_dependent_process.TimeDependentProcess` class.

        :ivar dict adjustment:  holding all state variables' adjustments
                                of the implicit process which are the
                                differences between the new states (which have
                                been solved through matrix inversion) and the
                                old states.

        """
        newstate = self._implicit_solver()
        adjustment = {}
        tendencies = {}
        for name, var in self.state.items():
            adjustment[name] = newstate[name] - var
            tendencies[name] = adjustment[name] / self.timestep
        # express the adjustment (already accounting for the finite time step)
        #  as a tendency per unit time, so that it can be applied along with explicit
        self.adjustment = adjustment
        self._update_diagnostics(newstate)
        return tendencies

    def _update_diagnostics(self, newstate):
        '''This method is called each timestep after the new state is computed
        with the implicit solver. Daughter classes can implement this method to
        compute any diagnostic quantities using the new state.'''
        pass



'''Tests for the 1D advection-diffusion solver against an analytical benchmark.'''
from __future__ import division
import numpy as np
import pytest
from climlab.dynamics import adv_diff_numerics

L = 1E7  # meters
J = 50
x0 = 0.
xJ = L
timestep = 3600.*24.
Psi0 = 1.  # some psi units
U0 = 10.  # m/s
Uscale_mixing = 10.  # m/s
Lscale_mixing = 1E5  # meters
Kconst = Uscale_mixing * Lscale_mixing
offset = 2E6

def flux_analytical(Xb, Psi0, U0, Kconst, L):
    return Psi0*np.sin(np.pi*Xb/L)*(U0*np.sin(np.pi*Xb/L)**2-2*Kconst*np.pi/L*np.cos(np.pi*Xb/L))

def tendency_analytical(X, Psi0, U0, Kconst, L):
    return (-Psi0*np.pi/L*(3*U0*np.sin(np.pi*X/L)**2*np.cos(np.pi*X/L)
                - 2*Kconst*np.pi/L*(np.cos(np.pi*X/L)**2-np.sin(np.pi*X/L)**2)))

@pytest.mark.fast
def test_uniform_grid():
    Xb = np.linspace(x0, xJ, J+1)
    X = np.linspace(x0+(Xb[1]-Xb[0])/2, xJ - (Xb[-1]-Xb[-2])/2, J)
    prescribed_flux = np.zeros_like(Xb)
    prescribed_source = np.zeros_like(X)
    Wb = np.ones_like(Xb)
    W = np.ones_like(X)
    Psi = Psi0 * (np.sin(np.pi*X/L))**2
    U = U0 * np.sin(np.pi*Xb/L)
    K = np.ones_like(Xb) * Kconst
    source = adv_diff_numerics.compute_source(X,Xb,prescribed_flux,prescribed_source,W,Wb)
    tridiag = adv_diff_numerics.advdiff_tridiag(X, Xb, K, U, W, Wb, use_banded_solver=True)
    F_analytical = flux_analytical(Xb, Psi0, U0, Kconst, L)
    F_numerical = adv_diff_numerics.total_flux(X, Xb, K, U, Psi, prescribed_flux)
    dPsidt_analytical = tendency_analytical(X, Psi0, U0, Kconst, L)
    dPsidt_numerical = adv_diff_numerics.compute_tendency(Psi, tridiag, source, use_banded_solver=True)

    assert F_numerical == pytest.approx(F_analytical, abs=0.01, rel=0.001)
    assert dPsidt_numerical == pytest.approx(dPsidt_analytical, abs=0.01, rel=0.001)
    adv_diff_numerics.implicit_step_forward(Psi, tridiag, source, timestep, use_banded_solver=True)

@pytest.mark.fast
def test_nonuniform_grid():
    Xb = np.geomspace(x0+offset, xJ+offset, J+1)-offset
    Xb[0] = 0.; Xb[-1] = xJ
    X = Xb[:-1] + (Xb[1:]-Xb[:-1])/2
    prescribed_flux = np.zeros_like(Xb)
    prescribed_source = np.zeros_like(X)
    Wb = np.ones_like(Xb)
    W = np.ones_like(X)
    Psi = Psi0 * (np.sin(np.pi*X/L))**2
    U = U0 * np.sin(np.pi*Xb/L)
    K = np.ones_like(Xb) * Kconst
    source = adv_diff_numerics.compute_source(X,Xb,prescribed_flux,prescribed_source,W,Wb)
    tridiag = adv_diff_numerics.advdiff_tridiag(X, Xb, K, U, W, Wb, use_banded_solver=False)
    F_analytical = flux_analytical(Xb, Psi0, U0, Kconst, L)
    F_numerical = adv_diff_numerics.total_flux(X, Xb, K, U, Psi, prescribed_flux)
    dPsidt_analytical = tendency_analytical(X, Psi0, U0, Kconst, L)
    dPsidt_numerical = adv_diff_numerics.compute_tendency(Psi, tridiag, source, use_banded_solver=False)
    assert F_numerical == pytest.approx(F_analytical, abs=0.01, rel=0.01)
    assert dPsidt_numerical == pytest.approx(dPsidt_analytical, abs=0.01, rel=0.01)
    adv_diff_numerics.implicit_step_forward(Psi, tridiag, source, timestep, use_banded_solver=False)

@pytest.mark.fast
def test_nonuniform_multidim():
    P=7; N = 4; M = 5  # arbitrary
    X = np.zeros((P,N,M,J))
    Xb = np.zeros((P,N,M,J+1))
    for p in range(P):
        for n in range(N):
            for m in range(M):
                Xb[p,n,m,:] = np.geomspace(x0+offset, xJ+offset, J+1)-offset
                Xb[p,n,m,0] = 0.; Xb[p,n,m,-1] = xJ
                X[p,n,m,:] = Xb[p,n,m,:-1] + (Xb[p,n,m,1:]-Xb[p,n,m,:-1])/2
    Psi = Psi0 * (np.sin(np.pi*X/L))**2
    K = np.ones_like(Xb) * Kconst
    U = U0 * np.sin(np.pi*Xb/L)
    prescribed_flux = np.zeros_like(U)
    prescribed_source = np.zeros_like(Psi)
    source = adv_diff_numerics.compute_source(X,Xb)
    tridiag = adv_diff_numerics.advdiff_tridiag(X, Xb, K, U)
    F_analytical = flux_analytical(Xb, Psi0, U0, Kconst, L)
    F_numerical = adv_diff_numerics.total_flux(X, Xb, K, U, Psi)
    dPsidt_analytical = tendency_analytical(X, Psi0, U0, Kconst, L)
    dPsidt_numerical = adv_diff_numerics.compute_tendency(Psi, tridiag, source)
    assert F_numerical == pytest.approx(F_analytical, abs=0.01, rel=0.01)
    assert dPsidt_numerical == pytest.approx(dPsidt_analytical, abs=0.01, rel=0.01)
    adv_diff_numerics.implicit_step_forward(Psi, tridiag, source, timestep, use_banded_solver=False)



from __future__ import division, print_function, absolute_import
import numpy as np
from climlab import constants as const
from climlab.solar.insolation import daily_insolation
from climlab.solar.orbital import OrbitalTable
from climlab.solar.orbital.long import OrbitalTable as LongOrbitalTable
from climlab import EBM_seasonal
from climlab.solar.orbital_cycles import OrbitalCycles
from climlab.surface import StepFunctionAlbedo
import pytest


@pytest.mark.fast
def test_daily_insolation():
    lat = np.linspace(-90., 90., 500)
    days = np.linspace(0, const.days_per_year, 365)
    Q = daily_insolation(lat, days)

    # check the range of Q
    np.testing.assert_almost_equal(Q.max(), 562.0333475)
    np.testing.assert_almost_equal(Q.min(), 0.0)

    # check the area integral
    Q_area_int = (np.sum(np.mean(Q, axis=1) * np.cos(np.deg2rad(lat))) /
                  np.sum(np.cos(np.deg2rad(lat))) )
    np.testing.assert_almost_equal(Q_area_int, 341.384184481)

@pytest.mark.fast
def test_orbital_parameters():
    kyears = np.arange( -1000., 1.)
    orb = OrbitalTable.interp(kyear=kyears)

    # check that orb has the right dictionary keys
    # key: (min, max)
    orb_expected = {'ecc': (0.004, 0.057),
                    #'long_peri': (2.3, 360),
                    'obliquity': (22, 24.5) }

    for k in orb_expected:
        assert orb[k].min() > orb_expected[k][0]
        assert orb[k].max() < orb_expected[k][1]

@pytest.mark.fast
def test_orbital_interpolation():
    # check to see if we get smooth results when we interpolate
    # orbital parameters at high temporal frequency
    kyears = np.linspace(-11, 0, 1001)
    orb = OrbitalTable.interp(kyear=kyears)
    S30 = daily_insolation(lat=30, day=172, orb=orb)
    # there should be no abrupt changes
    # test if insolation varies by more than 0.1 W/m2 per year
    assert S30.diff(dim='kyear').max() < 0.1


@pytest.mark.slow
def test_long_orbital_parameters():
    kyears = np.arange( -1000., +500.)
    orb = LongOrbitalTable.interp(kyear=kyears)

    # check that orb has the right dictionary keys
    # key: (min, max)
    orb_expected = {'ecc': (0.0018, 0.0579),
                    #'long_peri': (0.182, 360),
                    'obliquity': (22, 24.5) }

    for k in orb_expected:
        assert orb[k].min() > orb_expected[k][0]
        assert orb[k].max() < orb_expected[k][1]

#  Tests of automatic orbital cycles with EBM
@pytest.mark.slow
def test_orbital_cycles():
    ebm = EBM_seasonal()
    #  add an albedo feedback
    alb = StepFunctionAlbedo(state=ebm.state, **ebm.param)
    ebm.add_subprocess('albedo', alb)
    ebm.subprocess['SW'].albedo = alb.albedo
    #  run for 1,000 orbital years, but only 100 model years
    experiment = OrbitalCycles(ebm, kyear_start=-20, kyear_stop=-19,
                               orbital_year_factor=10.)
    assert experiment.orb_kyear == -20.
    np.testing.assert_almost_equal(experiment.T_segments_global, 11.48520525)



'''
Modules for radiative transfer in vertical columns,
along with processes for insolation and fixed relative humidity.
'''
from __future__ import absolute_import

from .aplusbt import AplusBT, AplusBT_CO2
from .absorbed_shorwave import SimpleAbsorbedShortwave
from .boltzmann import Boltzmann
from .insolation import FixedInsolation, P2Insolation, AnnualMeanInsolation, DailyInsolation
from .nband import NbandRadiation, ThreeBandSW
from .water_vapor import ManabeWaterVapor
#from radiation import Radiation, Radiation_SW, Radiation_LW
from .cam3 import CAM3, CAM3_LW, CAM3_SW
from .rrtm import RRTMG, RRTMG_LW, RRTMG_SW



from __future__ import division
import numpy as np
import climlab
from climlab.tests.xarray_test import to_xarray
import pytest

num_lev = 30
alb = 0.25

@pytest.fixture()
def rcm():
    # initial state (temperatures)
    state = climlab.column_state(num_lev=num_lev, num_lat=1, water_depth=5.)
    ## Create individual physical process models:
    #  fixed relative humidity
    h2o = climlab.radiation.ManabeWaterVapor(state=state, name='H2O')
    #  Hard convective adjustment
    convadj = climlab.convection.ConvectiveAdjustment(state=state, name='ConvectiveAdjustment',
                                                      adj_lapse_rate=6.5)
    # CAM3 radiation with default parameters and interactive water vapor
    rad = climlab.radiation.CAM3(state=state, albedo=alb, specific_humidity=h2o.q, name='Radiation')
    # Couple the models
    rcm = climlab.couple([h2o,convadj,rad], name='RCM')
    return rcm


@pytest.mark.compiled
@pytest.mark.fast
def test_rce(rcm):
    '''Test a single-column radiative-convective model with CAM3 radiation and
    fixed relative humidity.'''
    rcm.step_forward()
    # Did a diagnostic get properly updated?
    assert np.all(rcm.OLR == rcm.subprocess.Radiation.OLR)
    #rcm.integrate_years(5)
    #assert(np.isclose(rcm.Ts, ))
    # Test the xarray interface
    to_xarray(rcm)

@pytest.mark.compiled
@pytest.mark.slow
def test_re_radiative_forcing():
    state = climlab.column_state(num_lev=num_lev)
    rad = climlab.radiation.CAM3(state=state)
    rad.integrate_years(2)
    assert np.abs(rad.ASR - rad.OLR) < 0.1  # close to energy balance
    rad2 = climlab.process_like(rad)
    rad2.absorber_vmr['CO2'] *= 2.
    rad2.compute_diagnostics()
    assert (rad2.ASR - rad2.OLR) > 1.  # positive radiative forcing

@pytest.mark.compiled
@pytest.mark.slow
def test_rce_radiative_forcing(rcm):
    '''Run a single-column radiative-convective model with CAM3 radiation
    out to equilibrium. Clone the model, double CO2 and measure the instantaneous
    change in TOA flux. It should be positive net downward flux.'''
    rcm.integrate_years(5.)
    assert np.abs(rcm.ASR - rcm.OLR) < 0.1  # close to energy balance
    rcm2 = climlab.process_like(rcm)
    rcm2.subprocess['Radiation'].absorber_vmr['CO2'] *= 2.
    rcm2.compute_diagnostics()
    assert (rcm2.ASR - rcm2.OLR) > 1.  # positive radiative forcing

@pytest.mark.compiled
@pytest.mark.fast
def test_cam3_multidim():
    state = climlab.column_state(num_lev=40, num_lat=3, water_depth=5.)
    rad = climlab.radiation.CAM3(state=state)
    # Can we integrate?
    rad.step_forward()
    assert rad.OLR.shape == rad.Ts.shape



#  Trying a new data model for state variables and domains:
#  Create a new sub-class of numpy.ndarray
#  that has as an attribute the domain itself

# Following a tutorial on subclassing ndarray here:
#
# http://docs.scipy.org/doc/numpy/user/basics.subclassing.html
from __future__ import division
import numpy as np
from climlab.domain.xarray import Field_to_xarray


class Field(np.ndarray):
    """Custom class for climlab gridded quantities, called Field.

    This class behaves exactly like :py:class:`numpy.ndarray`
    but every object has an attribute called ``self.domain``
    which is the domain associated with that field (e.g. state variables).

    **Initialization parameters** \n

    An instance of ``Field`` is initialized with the following
    arguments:

    :param array input_array:   the array which the Field object should be
                                initialized with
    :param domain:              the domain associated with that field
                                (e.g. state variables)
    :type domain:               :class:`~climlab.domain.domain._Domain`

    **Object attributes** \n

    Following object attribute is generated during initialization:

    :var domain:               the domain associated with that field
                                (e.g. state variables)
    :vartype domain:            :class:`~climlab.domain.domain._Domain`


    :Example:

        ::

            >>> import climlab
            >>> import numpy as np
            >>> from climlab import domain
            >>> from climlab.domain import field

            >>> # distribution of state
            >>> distr = np.linspace(0., 10., 30)
            >>> # domain creation
            >>> sfc, atm = domain.single_column()
            >>> # build state of type Field
            >>> s = field.Field(distr, domain=atm)

            >>> print s
            [  0.           0.34482759   0.68965517   1.03448276   1.37931034
               1.72413793   2.06896552   2.4137931    2.75862069   3.10344828
               3.44827586   3.79310345   4.13793103   4.48275862   4.82758621
               5.17241379   5.51724138   5.86206897   6.20689655   6.55172414
               6.89655172   7.24137931   7.5862069    7.93103448   8.27586207
               8.62068966   8.96551724   9.31034483   9.65517241  10.        ]

            >>> print s.domain
            climlab Domain object with domain_type=atm and shape=(30,)

            >>> # can slice this and it preserves the domain
            >>> #  a more full-featured implementation would have intelligent
            >>> #  slicing like in iris
            >>> s.shape == s.domain.shape
            True
            >>> s[:1].shape == s[:1].domain.shape
            False

            >>> #  But some things work very well. E.g. new field creation:
            >>> s2 = np.zeros_like(s)

            >>> print s2
            [ 0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.
              0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.  0.]

            >>> print s2.domain
            climlab Domain object with domain_type=atm and shape=(30,)

    """
    def __new__(cls, input_array, domain=None, interfaces=False):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        #obj = np.asarray(input_array).view(cls)
        # This should ensure that shape is (1,) for scalar input
        #obj = np.atleast_1d(input_array).view(cls)
        # add the new attribute to the created instance
        #  do some checking for correct dimensions

        # input argument interfaces indicates whether input_array exists
        # on cell interfaces for each dimensions
        #  It should be either a single Boolean
        #  or an array of Booleans compatible with number of dimensions
        if input_array is None:
            return None
        else:
            try:
                shape = np.array(domain.shape) + np.where(interfaces,1,0)
            except:
                raise ValueError('domain and interfaces inconsistent.')
        try:
            #assert obj.shape == domain.shape
            #  This will work if input_array is any of:
            #   - scalar
            #   - same shape as domain
            #   - broadcast-compatible with domain shape
            obj = (input_array * np.ones(shape)).view(cls)
            assert np.all(obj.shape == shape)
        except:
            try:
                # Do we get a match if we add a singleton dimension
                #  (e.g. a singleton depth axis)?
                obj = np.expand_dims(input_array, axis=-1).view(cls)
                assert np.all(obj.shape == shape)
                #obj = np.transpose(np.atleast_2d(obj))
                #if obj.shape == domain.shape:
                #    obj.domain = domain
            except:
                raise ValueError('Cannot reconcile shapes of input_array and domain.')
        obj.domain = domain
        obj.interfaces = interfaces
        #  would be nice to have some automatic domain creation here if none given

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(Field, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. Field():
        #    obj is None
        #    (we're in the middle of the Field.__new__
        #    constructor, and self.domain will be set when we return to
        #    Field.__new__)
        if obj is None: return
        # From view casting - e.g arr.view(Field):
        #    obj is arr
        #    (type(obj) can be Field)
        # From new-from-template - e.g statearr[:3]
        #    type(obj) is Field
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'domain', because this
        # method sees all creation of default objects - with the
        # Field.__new__ constructor, but also with
        # arr.view(Field).
        try:
            self.domain = obj.domain
        except:
            self.domain = None
        try:
            self.interfaces = obj.interfaces
        except:
            pass
        # We do not need to return anything

##  Loosely based on the approach in numpy.ma.core.MaskedArray
#   This determines how we slice a Field object
    def __getitem__(self, indx):
        """
        x.__getitem__(y) <==> x[y]
        Return the item described by i, as a Field.
        """
        #  create a view of just the data as np.ndarray and slice it
        dout = self.view(np.ndarray)[indx]
        try:
            #Force dout to type Field
            dout = dout.view(type(self))
            # Now slice the domain
            dout.domain = self.domain[indx]
            # Inherit attributes from self
            if hasattr(self, 'interfaces'):
                dout.interfaces = self.interfaces
        except:
            # The above will fail if we extract a single item
            # in which case we should just return the item
            pass
        return dout

    def to_xarray(self):
        """Convert Field object to xarray.DataArray"""
        return Field_to_xarray(self)


def global_mean(field):
    """Calculates the latitude weighted global mean of a field
    with latitude dependence.

    :param Field field: input field
    :raises: :exc:`ValueError` if input field has no latitude axis
    :return: latitude weighted global mean of the field
    :rtype: float

    :Example:

        initial global mean temperature of EBM model::

            >>> import climlab
            >>> model = climlab.EBM()
            >>> climlab.global_mean(model.Ts)
            Field(11.997968598413685)

    """
    try:
        lat = field.domain.lat.points
    except:
        raise ValueError('No latitude axis in input field.')
    try:
        #  Field is 2D latitude / longitude
        lon = field.domain.lon.points
        return _global_mean_latlon(field.squeeze())
    except:
        #  Field is 1D latitude only (zonal average)
        lat_radians = np.deg2rad(lat)
        return _global_mean(field.squeeze(), lat_radians)


def _global_mean(array, lat_radians):
    #  Use np.array() here to strip the Field data and return a plain array
    #  (This will be more graceful once we are using xarray.DataArray
    #  for all internal grid info instead of the Field object)
    return np.array(np.average(array, weights=np.cos(lat_radians)))


def _global_mean_latlon(field):
    dom = field.domain
    lon, lat = np.meshgrid(dom.lon.points, dom.lat.points)
    dy = np.deg2rad(np.diff(dom.lat.bounds))
    dx = np.deg2rad(np.diff(dom.lon.bounds))*np.cos(np.deg2rad(lat))
    area = dx * dy[:,np.newaxis]  # grid cell area in radians^2
    return np.array(np.average(field, weights=area))


def to_latlon(array, domain, axis = 'lon'):
    """Broadcasts a 1D axis dependent array across another axis.

    :param array input_array:   the 1D array used for broadcasting
    :param domain:              the domain associated with that
                                array
    :param axis:                the axis that the input array will
                                be broadcasted across
                                [default: 'lon']
    :return:                    Field with the same shape as the
                                domain
    :Example:

        ::

            >>> import climlab
            >>> from climlab.domain.field import to_latlon
            >>> import numpy as np

            >>> state = climlab.surface_state(num_lat=3, num_lon=4)
            >>> m = climlab.EBM_annual(state=state)
            >>> insolation = np.array([237., 417., 237.])
            >>> insolation = to_latlon(insolation, domain = m.domains['Ts'])
            >>> insolation.shape
            (3, 4, 1)
            >>> insolation
            Field([[[ 237.],   [[ 417.],   [[ 237.],
                    [ 237.],    [ 417.],    [ 237.],
                    [ 237.],    [ 417.],    [ 237.],
                    [ 237.]],   [ 417.]],   [ 237.]]])

    """
    #  if array is latitude dependent (has the same shape as lat)
    axis, array, depth = np.meshgrid(domain.axes[axis].points, array,
                                    domain.axes['depth'].points)
    if axis == 'lat':
    #  if array is longitude dependent (has the same shape as lon)
        np.swapaxes(array,1,0)
    return Field(array, domain=domain)



from __future__ import division
from builtins import str
from builtins import object
from climlab.domain.axis import Axis
from climlab.utils import heat_capacity


class _Domain(object):
    """Private parent class for `Domains`.

    A `Domain` defines an area or spatial base for a climlab
    :class:`~climlab.process.process.Process` object. It consists of axes which
    are :class:`~climlab.domain.axis.Axis` objects that define the dimensions
    of the `Domain`.

    In a `Domain` the heat capacity of grid points, bounds or cells/boxes is
    specified.

    There are daughter classes :class:`~climlab.domain.domain.Atmosphere` and
    :class:`~climlab.domain.domain.Ocean` of the private
    :class:`~climlab.domain.domain._Domain` class implemented which themselves
    have daughter classes :class:`~climlab.domain.domain.SlabAtmosphere` and
    :class:`~climlab.domain.domain.SlabOcean`.

    Several methods are implemented that create `Domains` with special
    specifications. These are

	- :func:`~climlab.domain.domain.single_column`

	- :func:`~climlab.domain.domain.zonal_mean_column`

	- :func:`~climlab.domain.domain.box_model_domain`


    **Initialization parameters** \n

    An instance of ``_Domain`` is initialized with the following
    arguments:

    :param axes:    Axis object or dictionary of Axis object where domain will
                    be defined on.
    :type axes:     dict or :class:`~climlab.domain.axis.Axis`


    **Object attributes** \n

    Following object attributes are generated during initialization:

    :ivar str domain_type:      Set to ``'undefined'``.
    :ivar dict axes:            A dictionary of the domains axes. Created by
                                :func:`_make_axes_dict` called with input
                                argument ``axes``
    :ivar int numdims:          Number of :class:`~climlab.domain.axis.Axis` objects
                                in ``self.axes`` dictionary.
    :ivar dict ax_index:        A dictionary of domain axes and their corresponding index
                                in an ordered list of the axes with:        \n
                                - ``'lev'`` or ``'depth'`` is last
                                - ``'lat'`` is second last
    :ivar tuple shape:          Number of points of all domain axes. Order in
                                tuple given by ``self.ax_index``.
    :ivar array heat_capacity:  the domain's heat capacity over axis specified
                                in function call of :func:`set_heat_capacity`

    """
    def __str__(self):
        return ("climlab Domain object with domain_type=" + self.domain_type + " and shape=" +
                str(self.shape))
    def __init__(self, axes=None, **kwargs):
        self.domain_type = 'undefined'
        # self.axes should be a dictionary of axes
        # make it possible to give just a single axis:
        self.axes = self._make_axes_dict(axes)
        self.numdims = len(list(self.axes.keys()))
        shape = []
        axcount = 0
        axindex = {}
        #  ordered list of axes
        #  lev OR depth is last
        #  lat is second-last
        add_lev = False
        add_depth = False
        add_lon = False
        add_lat = False
        axlist = list(self.axes.keys())
        if 'lev' in axlist:
            axlist.remove('lev')
            add_lev = True
        elif 'depth' in axlist:
            axlist.remove('depth')
            add_depth = True
        if 'lon' in axlist:
            axlist.remove('lon')
            add_lon = True
        if 'lat' in axlist:
            axlist.remove('lat')
            add_lat = True
        axlist2 = axlist[:]
        if add_lat:
            axlist2.append('lat')
        if add_lon:
            axlist2.append('lon')
        if add_depth:
            axlist2.append('depth')
        if add_lev:
            axlist2.append('lev')
        #for axType, ax in self.axes.iteritems():
        for axType in axlist2:
            ax = self.axes[axType]
            shape.append(ax.num_points)
            #  can access axes as object attributes
            setattr(self, axType, ax)
            #
            axindex[axType] = axcount
            axcount += 1
        self.axis_index = axindex
        self.shape = tuple(shape)

        self.set_heat_capacity()

    def set_heat_capacity(self):
        """A dummy function to set the heat capacity of a domain.

        *Should be overridden by daugter classes.*

        """
        self.heat_capacity = None
        #  implemented by daughter classes

    def _make_axes_dict(self, axes):
        """Makes an axes dictionary.

        .. note::

            In case the input is ``None``, the dictionary :code:`{'empty': None}`
            is returned.

        **Function-call argument** \n

        :param axes:    axes input
        :type axes:     dict or single instance of
                        :class:`~climlab.domain.axis.Axis` object or ``None``
        :raises: :exc:`ValueError`  if input is not an instance of Axis class
                                    or a dictionary of Axis objetcs
        :returns: dictionary of input axes
        :rtype: dict

        """
        if type(axes) is dict:
            axdict = axes
        elif type(axes) is Axis:
            ax = axes
            axdict = {ax.axis_type: ax}
        elif axes is None:
            axdict = {'empty': None}
        else:
            raise ValueError('axes needs to be Axis object or dictionary of Axis object')
        return axdict

    def __getitem__(self, indx):
        # Make domains sliceable
        # First create a bare domain object (without calling the __init__ method)
        dout = type(self).__new__(type(self))
        # inherit *most* of the attributes of self
        #  For now we are just slicing the heat capacity
        #  But would be great to have some logic for slicing axes
        #   I am not 100% percent clear on how all this works
        #   But for now we're just going to "try" to slice to avoid
        #   some failures
        for key, value in self.__dict__.items():
            if key == 'heat_capacity':
                try:
                    dout.heat_capacity = self.heat_capacity[indx]
                except:
                    dout.heat_capacity = self.heat_capacity
            elif key == 'shape':
                try:
                    dout.shape = self.heat_capacity[indx].shape
                except:
                    dout.shape = self.shape
            else:
                setattr(dout, key, value)
        return dout


class Atmosphere(_Domain):
    """Class for the implementation of an Atmosphere Domain.

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.domain.domain._Domain`
    the following object attribute is modified during initialization:

    :ivar str domain_type:    is set to ``'atm'``

    :Example:

        Setting up an Atmosphere Domain::

            >>> import climlab
            >>> atm_ax = climlab.domain.Axis(axis_type='pressure', num_points=10)
            >>> atm_domain = climlab.domain.Atmosphere(axes=atm_ax)

            >>> print atm_domain
            climlab Domain object with domain_type=atm and shape=(10,)

            >>> atm_domain.axes
            {'lev': <climlab.domain.axis.Axis object at 0x7fe5b8ef8e10>}

            >>> atm_domain.heat_capacity
            array([ 1024489.79591837,  1024489.79591837,  1024489.79591837,
                    1024489.79591837,  1024489.79591837,  1024489.79591837,
                    1024489.79591837,  1024489.79591837,  1024489.79591837,
                    1024489.79591837])

    """
    def __init__(self, **kwargs):
        super(Atmosphere, self).__init__(**kwargs)
        self.domain_type = 'atm'

    def set_heat_capacity(self):
        """Sets the heat capacity of the Atmosphere Domain.

        Calls the utils heat capacity function
        :func:`~climlab.utils.heat_capacity.atmosphere` and gives the delta
        array of grid points of it's level axis
        ``self.axes['lev'].delta`` as input.

        **Object attributes** \n

        During method execution following object attribute is modified:

        :ivar array heat_capacity:  the ocean domain's heat capacity over
                                    the ``'lev'`` Axis.

        """
        self.heat_capacity = heat_capacity.atmosphere(self.axes['lev'].delta)


class Ocean(_Domain):
    """Class for the implementation of an Ocean Domain.

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.domain.domain._Domain`
    the following object attribute is modified during initialization:

    :ivar str domain_type:    is set to ``'ocean'``

    :Example:

        Setting up an Ocean Domain::

            >>> import climlab
            >>> ocean_ax = climlab.domain.Axis(axis_type='depth', num_points=5)
            >>> ocean_domain = climlab.domain.Ocean(axes=ocean_ax)

            >>> print ocean_domain
            climlab Domain object with domain_type=ocean and shape=(5,)

            >>> ocean_domain.axes
            {'depth': <climlab.domain.axis.Axis object at 0x7fe5b8f102d0>}

            >>> ocean_domain.heat_capacity
            array([ 8362600.,  8362600.,  8362600.,  8362600.,  8362600.])

    """
    def __init__(self, **kwargs):
        super(Ocean, self).__init__(**kwargs)
        self.domain_type = 'ocean'

    def set_heat_capacity(self):
        """Sets the heat capacity of the Ocean Domain.

        Calls the utils heat capacity function
        :func:`~climlab.utils.heat_capacity.ocean` and gives the delta
        array of grid points of it's depth axis
        ``self.axes['depth'].delta`` as input.

        **Object attributes** \n

        During method execution following object attribute is modified:

        :ivar array heat_capacity:  the ocean domain's heat capacity over
                                    the ``'depth'`` Axis.

        """
        self.heat_capacity = heat_capacity.ocean(self.axes['depth'].delta)


def make_slabocean_axis(num_points=1):
    """Convenience method to create a simple axis for a slab ocean.

    **Function-call argument** \n

    :param int num_points:    number of points for the slabocean Axis [default: 1]
    :returns:               an Axis with ``axis_type='depth'`` and ``num_points=num_points``
    :rtype:                 :class:`~climlab.domain.axis.Axis`

    :Example:

        ::

            >>> import climlab
            >>> slab_ocean_axis = climlab.domain.make_slabocean_axis()

            >>> print slab_ocean_axis
            Axis of type depth with 1 points.

            >>> slab_ocean_axis.axis_type
            'depth'

            >>> slab_ocean_axis.bounds
            array([  0.,  10.])

            >>> slab_ocean_axis.units
            'meters'

    """
    depthax = Axis(axis_type='depth', num_points=num_points)
    return depthax

def make_slabatm_axis(num_points=1):
    """Convenience method to create a simple axis for a slab atmosphere.

    **Function-call argument** \n

    :param int num_points:   number of points for the slabatmosphere Axis [default: 1]
    :returns:               an Axis with ``axis_type='lev'`` and ``num_points=num_points``
    :rtype:                 :class:`~climlab.domain.axis.Axis`

    :Example:

        ::

            >>> import climlab
            >>> slab_atm_axis = climlab.domain.make_slabatm_axis()

            >>> print slab_atm_axis
            Axis of type lev with 1 points.

            >>> slab_atm_axis.axis_type
            'lev'

            >>> slab_atm_axis.bounds
            array([    0.,  1000.])

            >>> slab_atm_axis.units
            'mb'

    """
    depthax = Axis(axis_type='lev', num_points=num_points)
    return depthax



class SlabOcean(Ocean):
    """A class to create a SlabOcean Domain by default.

    Initializes the parent :class:`Ocean` class with a simple axis for a
    Slab Ocean created by :func:`make_slabocean_axis` which has just 1 cell
    in depth by default.

    :Example:

        Creating a SlabOcean Domain::

            >>> import climlab
            >>> slab_ocean_domain = climlab.domain.SlabOcean()

            >>> print slab_ocean_domain
            climlab Domain object with domain_type=ocean and shape=(1,)

            >>> slab_ocean_domain.axes
            {'depth': <climlab.domain.axis.Axis object at 0x7fe5c42814d0>}

            >>> slab_ocean_domain.heat_capacity
            array([ 41813000.])

    """
    def __init__(self, axes=make_slabocean_axis(), **kwargs):
        super(SlabOcean, self).__init__(axes=axes, **kwargs)

class SlabAtmosphere(Atmosphere):
    """A class to create a SlabAtmosphere Domain by default.

    Initializes the parent :class:`Atmosphere` class with a simple axis for a
    Slab Atmopshere created by :func:`make_slabatm_axis` which has just 1 cell
    in height by default.

    :Example:

        Creating a SlabAtmosphere Domain::

            >>> import climlab
            >>> slab_atm_domain = climlab.domain.SlabAtmosphere()

            >>> print slab_atm_domain
            climlab Domain object with domain_type=atm and shape=(1,)

            >>> slab_atm_domain.axes
            {'lev': <climlab.domain.axis.Axis object at 0x7fe5c4281610>}

            >>> slab_atm_domain.heat_capacity
            array([ 10244897.95918367])

    """
    def __init__(self, axes=make_slabatm_axis(), **kwargs):
        super(SlabAtmosphere, self).__init__(axes=axes, **kwargs)


def single_column(num_lev=30, water_depth=1., lev=None, **kwargs):
    """Creates domains for a single column of atmosphere overlying a slab of water.

    Can also pass a pressure array or pressure level axis object specified in ``lev``.

    If argument ``lev`` is not ``None`` then function tries to build a level axis
    and ``num_lev`` is ignored.

    **Function-call argument** \n

    :param int num_lev:         number of pressure levels
                                (evenly spaced from surface to TOA) [default: 30]
    :param float water_depth:   depth of the ocean slab [default: 1.]
    :param lev:                 specification for height axis (optional)
    :type lev:                  :class:`~climlab.domain.axis.Axis` or pressure array
    :raises: :exc:`ValueError`  if `lev` is given but neither Axis
                                nor pressure array.
    :returns:                   a list of 2 Domain objects (slab ocean, atmosphere)
    :rtype:                     :py:class:`list` of :class:`SlabOcean`, :class:`SlabAtmosphere`

    :Example:

        ::

            >>> from climlab import domain

            >>> sfc, atm = domain.single_column(num_lev=2, water_depth=10.)

            >>> print sfc
            climlab Domain object with domain_type=ocean and shape=(1,)

            >>> print atm
            climlab Domain object with domain_type=atm and shape=(2,)

    """
    if lev is None:
        levax = Axis(axis_type='lev', num_points=num_lev)
    elif isinstance(lev, Axis):
        levax = lev
    else:
        try:
            levax = Axis(axis_type='lev', points=lev)
        except:
            raise ValueError('lev must be Axis object or pressure array')
    depthax = Axis(axis_type='depth', bounds=[water_depth, 0.])
    slab = SlabOcean(axes=depthax, **kwargs)
    atm = Atmosphere(axes=levax, **kwargs)
    return slab, atm


def zonal_mean_surface(num_lat=90, water_depth=10., lat=None, **kwargs):
    """Creates a 1D slab ocean Domain in latitude with uniform water depth.

    Domain has a single heat capacity according to the specified water depth.

    **Function-call argument** \n

    :param int num_lat:         number of latitude points [default: 90]
    :param float water_depth:   depth of the slab ocean in meters [default: 10.]
    :param lat:                 specification for latitude axis (optional)
    :type lat:                  :class:`~climlab.domain.axis.Axis` or latitude array
    :raises: :exc:`ValueError`  if `lat` is given but neither Axis nor latitude array.
    :returns:                   surface domain
    :rtype:                     :class:`SlabOcean`

    :Example:

        ::

            >>> from climlab import domain
            >>> sfc = domain.zonal_mean_surface(num_lat=36)

            >>> print sfc
            climlab Domain object with domain_type=ocean and shape=(36, 1)

    """
    if lat is None:
        latax = Axis(axis_type='lat', num_points=num_lat)
    elif isinstance(lat, Axis):
        latax = lat
    else:
        try:
            latax = Axis(axis_type='lat', points=lat)
        except:
            raise ValueError('lat must be Axis object or latitude array')
    depthax = Axis(axis_type='depth', bounds=[water_depth, 0.])
    axes = {'depth': depthax, 'lat': latax}
    slab = SlabOcean(axes=axes, **kwargs)
    return slab

def surface_2D(num_lat=90, num_lon=180, water_depth=10., lon=None,
               lat=None, **kwargs):
    """Creates a 2D slab ocean Domain in latitude and longitude with uniform water depth.

    Domain has a single heat capacity according to the specified water depth.

    **Function-call argument** \n

    :param int num_lat:         number of latitude points [default: 90]
    :param int num_lon:         number of longitude points [default: 180]
    :param float water_depth:   depth of the slab ocean in meters [default: 10.]
    :param lat:                 specification for latitude axis (optional)
    :type lat:                  :class:`~climlab.domain.axis.Axis` or latitude array
    :param lon:                 specification for longitude axis (optional)
    :type lon:                  :class:`~climlab.domain.axis.Axis` or longitude array
    :raises: :exc:`ValueError`  if `lat` is given but neither Axis nor latitude array.
    :raises: :exc:`ValueError`  if `lon` is given but neither Axis nor longitude array.
    :returns:                   surface domain
    :rtype:                     :class:`SlabOcean`

    :Example:

        ::

            >>> from climlab import domain
            >>> sfc = domain.surface_2D(num_lat=36, num_lat=72)

            >>> print sfc
            climlab Domain object with domain_type=ocean and shape=(36, 72, 1)

    """
    if lat is None:
        latax = Axis(axis_type='lat', num_points=num_lat)
    elif isinstance(lat, Axis):
        latax = lat
    else:
        try:
            latax = Axis(axis_type='lat', points=lat)
        except:
            raise ValueError('lat must be Axis object or latitude array')
    if lon is None:
        lonax = Axis(axis_type='lon', num_points=num_lon)
    elif isinstance(lon, Axis):
        lonax = lon
    else:
        try:
            lonax = Axis(axis_type='lon', points=lon)
        except:
            raise ValueError('lon must be Axis object or longitude array')
    depthax = Axis(axis_type='depth', bounds=[water_depth, 0.])
    axes = {'lat': latax, 'lon': lonax, 'depth': depthax}
    slab = SlabOcean(axes=axes, **kwargs)
    return slab

def zonal_mean_column(num_lat=90, num_lev=30, water_depth=10., lat=None,
                      lev=None, **kwargs):
    """Creates two Domains with one water cell, a latitude axis and
    a level/height axis.

        * SlabOcean:    one water cell and a latitude axis above
          (similar to :func:`zonal_mean_surface`)
        * Atmosphere: a latitude axis and a level/height axis (two dimensional)


    **Function-call argument** \n

    :param int num_lat:         number of latitude points on the axis
                                [default: 90]
    :param int num_lev:         number of pressure levels
                                (evenly spaced from surface to TOA) [default: 30]
    :param float water_depth:   depth of the water cell (slab ocean) [default: 10.]
    :param lat:                 specification for latitude axis (optional)
    :type lat:                  :class:`~climlab.domain.axis.Axis` or latitude array
    :param lev:                 specification for height axis (optional)
    :type lev:                  :class:`~climlab.domain.axis.Axis` or pressure array
    :raises: :exc:`ValueError`  if `lat` is given but neither Axis nor latitude array.
    :raises: :exc:`ValueError`  if `lev` is given but neither Axis nor pressure array.
    :returns:                   a list of 2 Domain objects (slab ocean, atmosphere)
    :rtype:                     :py:class:`list` of :class:`SlabOcean`, :class:`Atmosphere`

    :Example:

        ::

            >>> from climlab import domain
            >>> sfc, atm = domain.zonal_mean_column(num_lat=36,num_lev=10)

            >>> print sfc
            climlab Domain object with domain_type=ocean and shape=(36, 1)

            >>> print atm
            climlab Domain object with domain_type=atm and shape=(36, 10)


    """
    if lat is None:
        latax = Axis(axis_type='lat', num_points=num_lat)
    elif isinstance(lat, Axis):
        latax = lat
    else:
        try:
            latax = Axis(axis_type='lat', points=lat)
        except:
            raise ValueError('lat must be Axis object or latitude array')
    if lev is None:
        levax = Axis(axis_type='lev', num_points=num_lev)
    elif isinstance(lev, Axis):
        levax = lev
    else:
        try:
            levax = Axis(axis_type='lev', points=lev)
        except:
            raise ValueError('lev must be Axis object or pressure array')

    depthax = Axis(axis_type='depth', bounds=[water_depth, 0.])
    #axes = {'depth': depthax, 'lat': latax, 'lev': levax}
    slab = SlabOcean(axes={'lat':latax, 'depth':depthax}, **kwargs)
    atm = Atmosphere(axes={'lat':latax, 'lev':levax}, **kwargs)
    return slab, atm

def box_model_domain(num_points=2, **kwargs):
    """Creates a box model domain (a single abstract axis).

    :param int num_points:      number of boxes [default: 2]
    :returns:                   Domain with single axis of type ``'abstract'``
                                and ``self.domain_type = 'box'``
    :rtype:                     :class:`_Domain`

    :Example:

        ::

            >>> from climlab import domain
            >>> box = domain.box_model_domain(num_points=2)

            >>> print box
            climlab Domain object with domain_type=box and shape=(2,)

    """
    ax = Axis(axis_type='abstract', num_points=num_points)
    boxes = _Domain(axes=ax, **kwargs)
    boxes.domain_type = 'box'
    return boxes



'''A collection of general purpose utility functions.'''

# Root URL for the climlab data repository
_datapath_http = "http://www.atmos.albany.edu/facstaff/brose/resources/climlab_data/"
# Alternate path that should also work
_threddspath_http = "http://thredds.atmos.albany.edu:8080/thredds/fileServer/CLIMLAB/"



from __future__ import division
from builtins import str
from builtins import object
import numpy as np
from climlab import constants as const


axis_types = ['lev', 'lat', 'lon', 'depth', 'abstract']


# will need to implement a simple cartesian distance axis type
# and probaly also an abstract dimensionless axis type (for box models)

class Axis(object):
    """Creates a new climlab Axis object.

    An :class:`~climlab.domain.axis.Axis` is an object where information of a
    spacial dimension of a :class:`~climlab.domain.domain._Domain` are specified.

    These include the `type` of the axis, the `number of points`, location of
    `points` and `bounds` on the spatial dimension, magnitude of bounds
    differences `delta` as well as their `unit`.

    The `axes` of a :class:`~climlab.domain.domain._Domain` are stored in the
    dictionary axes, so they can be accessed through ``dom.axes`` if ``dom``
    is an instance of :class:`~climlab.domain.domain._Domain`.


    **Initialization parameters** \n

    An instance of ``Axis`` is initialized with the following
    arguments *(for detailed information see Object attributes below)*:

    :param str axis_type:   information about the type of axis
                            [default: 'abstract']
    :param int num_points:  number of points on axis
                            [default: 10]
    :param array points:    array with specific points (optional)
    :param array bounds:    array with specific bounds between points (optional)
    :raises: :exc:`ValueError`
                            if ``axis_type`` is not one of the valid types or
                            their euqivalents (see below).
    :raises: :exc:`ValueError`
                            if ``points`` are given and not array-like.
    :raises: :exc:`ValueError`
                            if ``bounds`` are given and not array-like.

    **Object attributes** \n

    Following object attributes are generated during initialization:

    :ivar str axis_type:    Information about the type of axis. Valid axis types are:

                                * ``'lev'``
                                * ``'lat'``
                                * ``'lon'``
                                * ``'depth'``
                                * ``'abstract'`` (default)

    :ivar int num_points:   number of points on axis
    :ivar str units:        Unit of the axis. During intialization the unit is
                            chosen from the ``defaultUnits`` dictionary (see below).
    :ivar array points:     array with all points of the axis (grid)
    :ivar array bounds:     array with all bounds between points (staggered grid)
    :ivar array delta:      array with spatial differences between bounds


    **Axis Types** \n

    A couple of differing axis type strings are rendered to valid axis types.
    Alternate forms are listed here:

    * ``'lev'``
        * ``'p'``
        * ``'press'``
        * ``'pressure'``
        * ``'P'``
        * ``'Pressure'``
        * ``'Press'``
    * ``'lat'``
        * ``'Latitude'``
        * ``'latitude'``
    * ``'lon'``
        * ``'Longitude'``
        * ``'longitude'``
    * ``'depth'``
        * ``'Depth'``
        * ``'waterDepth'``
        * ``'water_depth'``
        * ``'slab'``


    The **default units** are::

        defaultUnits = {'lev': 'mb',
                        'lat': 'degrees',
                        'lon': 'degrees',
                        'depth': 'meters',
                        'abstract': 'none'}

    If bounds are not given during initialization, **default end points**
    are used::

        defaultEndPoints = {'lev': (0., climlab.constants.ps),
                            'lat': (-90., 90.),
                            'lon': (0., 360.),
                            'depth': (0., 10.),
                            'abstract': (0, num_points)}

    :Example:

        Creation of a standalone Axis::

            >>> import climlab
            >>> ax = climlab.domain.Axis(axis_type='Latitude', num_points=36)

            >>> print ax
            Axis of type lat with 36 points.

            >>> ax.points
            array([-87.5, -82.5, -77.5, -72.5, -67.5, -62.5, -57.5, -52.5, -47.5,
                   -42.5, -37.5, -32.5, -27.5, -22.5, -17.5, -12.5,  -7.5,  -2.5,
                     2.5,   7.5,  12.5,  17.5,  22.5,  27.5,  32.5,  37.5,  42.5,
                    47.5,  52.5,  57.5,  62.5,  67.5,  72.5,  77.5,  82.5,  87.5])

            >>> ax.bounds
            array([-90., -85., -80., -75., -70., -65., -60., -55., -50., -45., -40.,
                   -35., -30., -25., -20., -15., -10.,  -5.,   0.,   5.,  10.,  15.,
                    20.,  25.,  30.,  35.,  40.,  45.,  50.,  55.,  60.,  65.,  70.,
                    75.,  80.,  85.,  90.])

            >>> ax.delta
            array([ 5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,
                    5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,
                    5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.])

    """
    def __str__(self):
        return ("Axis of type " + self.axis_type + " with " +
                str(self.num_points) + " points.")

    def __init__(self, axis_type='abstract', num_points=10, points=None, bounds=None):
        if axis_type in axis_types:
            pass
        elif axis_type in ['p', 'press', 'pressure', 'P', 'Pressure', 'Press']:
            axis_type = 'lev'
        elif axis_type in ['Latitude', 'latitude']:
            axis_type = 'lat'
        elif axis_type in ['Longitude', 'longitude']:
            axis_type = 'lon'
        elif axis_type in ['depth', 'Depth', 'waterDepth', 'water_depth', 'slab']:
            axis_type = 'depth'
        else:
            raise ValueError('axis_type %s not recognized' % axis_type)
        self.axis_type = axis_type

        defaultEndPoints = {'lev': (0., const.ps),
                            'lat': (-90., 90.),
                            'lon': (0., 360.),
                            'depth': (0., 10.),
                            'abstract': (0, num_points)}
        defaultUnits = {'lev': 'mb',
                        'lat': 'degrees',
                        'lon': 'degrees',
                        'depth': 'meters',
                        'abstract': 'none'}
        # if points and/or bounds are supplied, make sure they are increasing
        if points is not None:
            try:
                # using np.atleast_1d() ensures that we can use a single point
                points = np.sort(np.atleast_1d(np.array(points, dtype=float)))
            except:
                raise ValueError('points must be array_like.')
        if bounds is not None:
            try:
                bounds = np.sort(np.atleast_1d(np.array(bounds, dtype=float)))
            except:
                raise ValueError('bounds must be array_like.')

        if bounds is None:
            # assume default end points
            end0 = defaultEndPoints[axis_type][0]
            end1 = defaultEndPoints[axis_type][1]
            if points is not None:
                # only points are given
                num_points = points.size
                bounds = points[:-1] + np.diff(points)/2.
                temp = np.append(np.flipud(bounds), end0)
                bounds = np.append(np.flipud(temp), end1)
            else:
                # no points or bounds
                # create an evenly spaced axis
                delta = (end1 - end0) / num_points
                bounds = np.linspace(end0, end1, num_points+1)
                points = np.linspace(end0 + delta/2., end1-delta/2., num_points)
        else:  # bounds are given
            end0 = bounds[0]
            end1 = bounds[1]
            num_points = bounds.size - 1
            if points is None:
                # only bounds given. Assume points are halfway between bounds
                points = bounds[:-1] + np.diff(bounds)/2.
            else:
                # points and bounds both given, check that they are compatible
                if points.size != num_points:
                    raise ValueError('points and bounds have incompatible sizes')
        self.num_points = num_points
        self.units = defaultUnits[axis_type]
        # pressure axis should decrease from surface to TOA
        #  NO! Now define the lowest (near-to-surface) element as lev[-1]
        #   and the nearest to space as lev[0]
        #if axis_type is 'lev':
        #    points = np.flipud(points)
        #    bounds = np.flipud(bounds)
        self.points = points
        self.bounds = bounds
        self.delta = np.abs(np.diff(self.bounds))
